# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/acousticbrainz/ui_options_acousticbrainz_tags.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AcousticBrainzOptionsPage(object):
    def setupUi(self, AcousticBrainzOptionsPage):
        AcousticBrainzOptionsPage.setObjectName("AcousticBrainzOptionsPage")
        AcousticBrainzOptionsPage.setEnabled(True)
        AcousticBrainzOptionsPage.resize(527, 443)
        AcousticBrainzOptionsPage.setWindowTitle("")
        self.verticalLayout = QtWidgets.QVBoxLayout(AcousticBrainzOptionsPage)
        self.verticalLayout.setObjectName("verticalLayout")
        self.acousticbrainzTags_groupBox = QtWidgets.QGroupBox(AcousticBrainzOptionsPage)
        self.acousticbrainzTags_groupBox.setObjectName("acousticbrainzTags_groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.acousticbrainzTags_groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.add_simplemood = QtWidgets.QCheckBox(self.acousticbrainzTags_groupBox)
        self.add_simplemood.setObjectName("add_simplemood")
        self.verticalLayout_2.addWidget(self.add_simplemood)
        self.simplemood_tagname_label = QtWidgets.QLabel(self.acousticbrainzTags_groupBox)
        self.simplemood_tagname_label.setObjectName("simplemood_tagname_label")
        self.verticalLayout_2.addWidget(self.simplemood_tagname_label)
        self.simplemood_tagname = QtWidgets.QLineEdit(self.acousticbrainzTags_groupBox)
        self.simplemood_tagname.setObjectName("simplemood_tagname")
        self.verticalLayout_2.addWidget(self.simplemood_tagname)
        self.add_simplegenre = QtWidgets.QCheckBox(self.acousticbrainzTags_groupBox)
        self.add_simplegenre.setObjectName("add_simplegenre")
        self.verticalLayout_2.addWidget(self.add_simplegenre)
        self.simplegenre_tagname_label = QtWidgets.QLabel(self.acousticbrainzTags_groupBox)
        self.simplegenre_tagname_label.setObjectName("simplegenre_tagname_label")
        self.verticalLayout_2.addWidget(self.simplegenre_tagname_label)
        self.simplegenre_tagname = QtWidgets.QLineEdit(self.acousticbrainzTags_groupBox)
        self.simplegenre_tagname.setObjectName("simplegenre_tagname")
        self.verticalLayout_2.addWidget(self.simplegenre_tagname)
        self.add_keybpm = QtWidgets.QCheckBox(self.acousticbrainzTags_groupBox)
        self.add_keybpm.setObjectName("add_keybpm")
        self.verticalLayout_2.addWidget(self.add_keybpm)
        self.add_fullhighlevel = QtWidgets.QCheckBox(self.acousticbrainzTags_groupBox)
        self.add_fullhighlevel.setObjectName("add_fullhighlevel")
        self.verticalLayout_2.addWidget(self.add_fullhighlevel)
        self.add_sublowlevel = QtWidgets.QCheckBox(self.acousticbrainzTags_groupBox)
        self.add_sublowlevel.setObjectName("add_sublowlevel")
        self.verticalLayout_2.addWidget(self.add_sublowlevel)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        self.verticalLayout_2.addItem(spacerItem)
        self.sublowlevel_descr = QtWidgets.QLabel(self.acousticbrainzTags_groupBox)
        self.sublowlevel_descr.setTextFormat(QtCore.Qt.RichText)
        self.sublowlevel_descr.setObjectName("sublowlevel_descr")
        self.verticalLayout_2.addWidget(self.sublowlevel_descr)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem1)
        self.verticalLayout.addWidget(self.acousticbrainzTags_groupBox)

        self.retranslateUi(AcousticBrainzOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(AcousticBrainzOptionsPage)

    def retranslateUi(self, AcousticBrainzOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.acousticbrainzTags_groupBox.setTitle(_translate("AcousticBrainzOptionsPage", "AcousticBrainz Tags"))
        self.add_simplemood.setText(_translate("AcousticBrainzOptionsPage", "Add simple Mood tags"))
        self.simplemood_tagname_label.setText(_translate("AcousticBrainzOptionsPage", "Mood tag name:"))
        self.add_simplegenre.setText(_translate("AcousticBrainzOptionsPage", "Add simple Genre tags"))
        self.simplegenre_tagname_label.setText(_translate("AcousticBrainzOptionsPage", "Genre tag name:"))
        self.add_keybpm.setText(_translate("AcousticBrainzOptionsPage", "Add simple BPM and Key tags"))
        self.add_fullhighlevel.setText(_translate("AcousticBrainzOptionsPage", "Add all highlevel AcousticBrainz tags"))
        self.add_sublowlevel.setText(_translate("AcousticBrainzOptionsPage", "Add a subset of the lowlevel AcousticBrainz tags"))
        self.sublowlevel_descr.setText(_translate("AcousticBrainzOptionsPage", "<html><head/><body><p>The low level subset include:</p><ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" margin-top:12px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">rhythm:bpm</li><li style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">tonal:chords_change_rate</li><li style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">tonal:chords_key</li><li style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">tonal:chords_scale</li><li style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">tonal:key_key</li><li style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">tonal:key_scale</li></ul></body></html>"))



# MusicBrainz Picard plugin to add an ~instruments tag.
# Copyright (C) 2019  David Mandelberg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

PLUGIN_NAME = 'Instruments'
PLUGIN_AUTHOR = 'David Mandelberg'
PLUGIN_DESCRIPTION = """\
  Adds a multi-valued tag (~instruments) containing all the instruments (including vocals), 
  for use in scripts.
  """
PLUGIN_VERSION = '1.0.1'
PLUGIN_API_VERSIONS = ['2.0']
PLUGIN_LICENSE = 'GPL-3.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl-3.0.html'

from typing import Generator, Optional

from picard import metadata
from picard import plugin


def _iterate_instruments(instrument_list: str) -> Generator[str, None, None]:
  """Yields individual instruments from a string listing them.

  Args:
    instrument_list: List of instruments in the form 'A, B and C'.
  """
  remaining = instrument_list
  while remaining:
    instrument, _, remaining = remaining.partition(', ')
    if not remaining:
      instrument, _, remaining = instrument.partition(' and ')
      if ' and ' in remaining:
        raise ValueError('Instrument list contains multiple \'and\'s: {!r}'
                         .format(instrument_list))
    yield instrument


def _strip_instrument_prefixes(instrument: str) -> Optional[str]:
  """Returns the instrument name without qualifying prefixes, or None.

  Args:
    instrument: Potentially prefixed instrument name, e.g., 'solo bassoon'.

  Returns:
    The instrument name with all prefixes stripped, or None if there's nothing
    other than prefixes. The all-prefixes case can happen with relationships
    like 'guest performer'.
  """
  instrument_prefixes = {
      'additional',
      'guest',
      'solo',
  }
  remaining = instrument
  while remaining:
    prefix, sep, remaining = remaining.partition(' ')
    if prefix not in instrument_prefixes:
      return ''.join((prefix, sep, remaining))
  return None


def add_instruments(tagger, metadata_, *args):
  key_prefix = 'performer:'
  instruments = set()
  for key in metadata_.keys():
    if not key.startswith(key_prefix):
      continue
    for instrument in _iterate_instruments(key[len(key_prefix):]):
      instrument = _strip_instrument_prefixes(instrument)
      if instrument:
        instruments.add(instrument)
  metadata_['~instruments'] = list(instruments)


metadata.register_track_metadata_processor(
    add_instruments, priority=plugin.PluginPriority.HIGH)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Decade function'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = ('Add a $decade(date) function to get the decade '
                      'from a year. E.g. $decade(1994-04-05) will give "90s". '
                      'By default decades between 1920 and 2000 will be '
                      'shortened to two digits. You can disable this with '
                      'setting the second parameter to 0, e.g. '
                      '$decade(1994,0) will give "1990s".')
PLUGIN_VERSION = "1.0"
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2", "2.3"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"


from picard.script import register_script_function


def decade(date, shorten=True):
    """
    >>> decade("1970-09-18")
    '70s'
    >>> decade("1994")
    '90s'
    >>> decade("1994-04")
    '90s'
    >>> decade("1994-04-05")
    '90s'
    >>> decade("1994-04-05", shorten=False)
    '1990s'
    >>> decade("1901")
    '1900s'
    >>> decade("1917-08-22")
    '1910s'
    >>> decade("1921-10-10")
    '20s'
    >>> decade("1770-12-16")
    '1770s'
    >>> decade("2017-07-20")
    '2010s'
    >>> decade("2020")
    '2020s'
    >>> decade("992")
    '990s'
    >>> decade("992-01")
    '990s'
    >>> decade("")
    ''
    >>> decade(None)
    ''
    >>> decade("foo")
    ''
    """
    try:
        year = int(date.split('-')[0])
    except (AttributeError, ValueError):
        return ""
    decade = year // 10 * 10
    if shorten and 1920 <= decade < 2000:
        decade -= 1900
    return "%ds" % decade


def script_decade(parser, value, shorten=True):
    return decade(value, shorten and shorten != '0')


register_script_function(script_decade, name="decade")



# -*- coding: utf-8 -*-

PLUGIN_NAME = 'Album Artist Website'
PLUGIN_AUTHOR = 'Sophist, Sambhav Kothari, Philipp Wolfer'
PLUGIN_DESCRIPTION = '''Add's the album artist(s) Official Homepage(s)
(if they are defined in the MusicBrainz database).'''
PLUGIN_VERSION = '1.1'
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2"]
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard import config, log
from picard.util import LockableObject
from picard.metadata import register_track_metadata_processor
from functools import partial

class AlbumArtistWebsite:

    class ArtistWebsiteQueue(LockableObject):

        def __init__(self):
            LockableObject.__init__(self)
            self.queue = {}

        def __contains__(self, name):
            return name in self.queue

        def __iter__(self):
            return self.queue.__iter__()

        def __getitem__(self, name):
            self.lock_for_read()
            value = self.queue[name] if name in self.queue else None
            self.unlock()
            return value

        def __setitem__(self, name, value):
            self.lock_for_write()
            self.queue[name] = value
            self.unlock()

        def append(self, name, value):
            self.lock_for_write()
            if name in self.queue:
                self.queue[name].append(value)
                value = False
            else:
                self.queue[name] = [value]
                value = True
            self.unlock()
            return value

        def remove(self, name):
            self.lock_for_write()
            value = None
            if name in self.queue:
                value = self.queue[name]
                del self.queue[name]
            self.unlock()
            return value

    def __init__(self):
        self.website_cache = {}
        self.website_queue = self.ArtistWebsiteQueue()

    def add_artist_website(self, album, track_metadata, track_node, release_node):
        albumArtistIds = track_metadata.getall('musicbrainz_albumartistid')
        for artistId in albumArtistIds:
            # Jump through hoops to get track object!!
            track = album._new_tracks[-1]
            if artistId in self.website_cache:
                if self.website_cache[artistId]:
                    self.add_websites_to_track(track, self.website_cache[artistId])
            else:
                self.website_add_track(album, track, artistId)

    def website_add_track(self, album, track, artistId):
        self.album_add_request(album)
        if self.website_queue.append(artistId, (track, album)):
            host = config.setting["server_host"]
            port = config.setting["server_port"]
            path = "/ws/2/%s/%s" % ('artist', artistId)
            queryargs = {"inc": "url-rels"}
            return album.tagger.webservice.get(host, port, path,
                        partial(self.website_process, artistId),
                                parse_response_type="xml", priority=True, important=False,
                                queryargs=queryargs)

    def website_process(self, artistId, response, reply, error):
        if error:
            log.error("%s: %r: Network error retrieving artist record", PLUGIN_NAME, artistId)
            tuples = self.website_queue.remove(artistId)
            for track, album in tuples:
                self.album_remove_request(album)
            return
        urls = self.artist_process_metadata(artistId, response)
        self.website_cache[artistId] = urls
        tuples = self.website_queue.remove(artistId)
        for track, album in tuples:
            self.add_websites_to_track(track, urls)
            self.album_remove_request(album)

    def add_websites_to_track(self, track, urls):
        tm = track.metadata
        websites = tm.getall('website')
        websites += urls
        websites.sort()
        tm['website'] = websites
        for file in track.iterfiles(True):
            file.metadata['website'] = websites

    def album_add_request(self, album):
        album._requests += 1

    def album_remove_request(self, album):
        album._requests -= 1
        album._finalize_loading(None)

    def artist_process_metadata(self, artistId, response):
        log.debug("%s: %r: Processing Artist record for official website urls: %r", PLUGIN_NAME, artistId, response)
        relations = self.artist_get_relations(response)
        if not relations:
            log.info("%s: %r: Artist does have any associated urls.", PLUGIN_NAME, artistId)
            return []

        urls = []
        for relation in relations:
            log.debug("%s: %r: Examining: %r", PLUGIN_NAME, artistId, relation)
            if 'type' in relation.attribs and relation.type == 'official homepage':
                if 'target' in relation.children and len(relation.target) > 0:
                    log.debug("%s: Adding artist url: %s", PLUGIN_NAME, relation.target[0].text)
                    urls.append(relation.target[0].text)
                else:
                    log.debug("%s: No url in relation: %r", PLUGIN_NAME, relation)

        if urls:
            log.info("%s: %r: Artist Official Homepages: %r", PLUGIN_NAME, artistId, urls)
        else:
            log.info("%s: %r: Artist does not have any official website urls.", PLUGIN_NAME, artistId)
        return sorted(urls)

    def artist_get_relations(self, response):
        log.debug("%s: artist_get_relations called", PLUGIN_NAME)
        if 'metadata' in response.children and len(response.metadata) > 0:
            if 'artist' in response.metadata[0].children and len(response.metadata[0].artist) > 0:
                if 'relation_list' in response.metadata[0].artist[0].children and len(response.metadata[0].artist[0].relation_list) > 0:
                    if 'relation' in response.metadata[0].artist[0].relation_list[0].children:
                        log.debug("%s: artist_get_relations returning: %r", PLUGIN_NAME, response.metadata[0].artist[0].relation_list[0].relation)
                        return response.metadata[0].artist[0].relation_list[0].relation
                    else:
                        log.debug("%s: artist_get_relations - no relation in relation_list", PLUGIN_NAME)
                else:
                    log.debug("%s: artist_get_relations - no relation_list in artist", PLUGIN_NAME)
            else:
                log.debug("%s: artist_get_relations - no artist in metadata", PLUGIN_NAME)
        else:
            log.debug("%s: artist_get_relations - no metadata in response", PLUGIN_NAME)
        return None


register_track_metadata_processor(AlbumArtistWebsite().add_artist_website)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/theaudiodb/ui_options_theaudiodb.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TheAudioDbOptionsPage(object):
    def setupUi(self, TheAudioDbOptionsPage):
        TheAudioDbOptionsPage.setObjectName("TheAudioDbOptionsPage")
        TheAudioDbOptionsPage.resize(442, 364)
        self.vboxlayout = QtWidgets.QVBoxLayout(TheAudioDbOptionsPage)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.groupBox = QtWidgets.QGroupBox(TheAudioDbOptionsPage)
        self.groupBox.setObjectName("groupBox")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.groupBox)
        self.vboxlayout1.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout1.setSpacing(2)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setTextFormat(QtCore.Qt.RichText)
        self.label.setWordWrap(True)
        self.label.setOpenExternalLinks(True)
        self.label.setObjectName("label")
        self.vboxlayout1.addWidget(self.label)
        self.vboxlayout.addWidget(self.groupBox)
        self.verticalGroupBox = QtWidgets.QGroupBox(TheAudioDbOptionsPage)
        self.verticalGroupBox.setObjectName("verticalGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.theaudiodb_cdart_use_always = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.theaudiodb_cdart_use_always.setObjectName("theaudiodb_cdart_use_always")
        self.verticalLayout.addWidget(self.theaudiodb_cdart_use_always)
        self.theaudiodb_cdart_use_if_no_albumcover = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.theaudiodb_cdart_use_if_no_albumcover.setObjectName("theaudiodb_cdart_use_if_no_albumcover")
        self.verticalLayout.addWidget(self.theaudiodb_cdart_use_if_no_albumcover)
        self.theaudiodb_cdart_use_never = QtWidgets.QRadioButton(self.verticalGroupBox)
        self.theaudiodb_cdart_use_never.setObjectName("theaudiodb_cdart_use_never")
        self.verticalLayout.addWidget(self.theaudiodb_cdart_use_never)
        self.vboxlayout.addWidget(self.verticalGroupBox)
        self.theaudiodb_use_high_quality = QtWidgets.QCheckBox(TheAudioDbOptionsPage)
        self.theaudiodb_use_high_quality.setObjectName("theaudiodb_use_high_quality")
        self.vboxlayout.addWidget(self.theaudiodb_use_high_quality)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem)

        self.retranslateUi(TheAudioDbOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(TheAudioDbOptionsPage)

    def retranslateUi(self, TheAudioDbOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.groupBox.setTitle(_translate("TheAudioDbOptionsPage", "TheAudioDB cover art"))
        self.label.setText(_translate("TheAudioDbOptionsPage", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /></head><body>\n"
"<p>This plugin loads cover art from <a href=\"https://www.theaudiodb.com\">TheAudioDB</a>. TheAudioDB is a community Database of audio artwork and data. Their content is only possible thanks to the hard work of volunteer editors. If you like this plugin, please consider <a href=\"https://www.theaudiodb.com/user_register_now.php\">registering</a> as an editor or supporting the TheAudioDB <a href=\"https://www.patreon.com/thedatadb\">patreon campaign</a>.</p></body></html>"))
        self.verticalGroupBox.setTitle(_translate("TheAudioDbOptionsPage", "Medium images"))
        self.theaudiodb_cdart_use_always.setText(_translate("TheAudioDbOptionsPage", "Always load medium images"))
        self.theaudiodb_cdart_use_if_no_albumcover.setText(_translate("TheAudioDbOptionsPage", "Load only if no front cover is available"))
        self.theaudiodb_cdart_use_never.setText(_translate("TheAudioDbOptionsPage", "Never load medium images"))
        self.theaudiodb_use_high_quality.setText(_translate("TheAudioDbOptionsPage", "Use high resolution images, if available"))



from PyQt5 import QtCore, QtWidgets


class Ui_MusixmatchOptionsPage(object):

    def setupUi(self, MusixmatchOptionsPage):
        MusixmatchOptionsPage.setObjectName("MusixmatchOptionsPage")
        MusixmatchOptionsPage.resize(305, 317)
        self.vboxlayout = QtWidgets.QVBoxLayout(MusixmatchOptionsPage)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.api_key_group = QtWidgets.QGroupBox(MusixmatchOptionsPage)
        self.api_key_group.setObjectName("api_key_group")
        self.vboxlayout2 = QtWidgets.QVBoxLayout(self.api_key_group)
        self.vboxlayout2.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout2.setSpacing(2)
        self.vboxlayout2.setObjectName("vboxlayout2")
        self.api_key_label = QtWidgets.QLabel(self.api_key_group)
        self.api_key_label.setObjectName("api_key_label")
        self.vboxlayout2.addWidget(self.api_key_label)
        self.api_key = QtWidgets.QLineEdit(self.api_key_group)
        self.api_key.setObjectName("api_key")
        self.vboxlayout2.addWidget(self.api_key)
        self.vboxlayout.addWidget(self.api_key_group)
        spacerItem = QtWidgets.QSpacerItem(263, 21, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem)

        self.retranslateUi(MusixmatchOptionsPage)
        QtCore.QMetaObject.connectSlotsByName(MusixmatchOptionsPage)

    def retranslateUi(self, MusixmatchOptionsPage):
        self.api_key_label.setText(_("Musixmatch API Key"))



# -*- coding: utf-8 -*-
# Critiquebrainz plugin for Picard
# Copyright (C) 2022  Tobias Sarner
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

PLUGIN_NAME = 'Critiquebrainz Review Comment'
PLUGIN_AUTHOR = 'Tobias Sarner'
PLUGIN_DESCRIPTION = '''Uses Critiquebrainz for comment as review or rating.

WARNING: Experimental plugin. All guarantees voided by use.

Example: Taylor Swift
Release: Midnights
 https://musicbrainz.org/release/e348fdd6-f73b-47fe-94c4-670bfee26a39 ,
 https://critiquebrainz.org/release-group/0dcc84fb-c592-46e9-ba92-a52bb44dd553

Recording:
 https://musicbrainz.org/recording/93113326-93e9-409c-a3d6-5ec91864ba30 ,
 https://critiquebrainz.org/recording/93113326-93e9-409c-a3d6-5ec91864ba30'''
PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"
PLUGIN_VERSION = "1.0.1"
PLUGIN_API_VERSIONS = ["2.0"]

from functools import partial
from picard import log
from picard.metadata import register_album_metadata_processor
from picard.metadata import register_track_metadata_processor
from picard.webservice import ratecontrol
from picard.util import load_json

CRITIQUEBRAINZ_HOST = "critiquebrainz.org"
CRITIQUEBRAINZ_PORT = 80

ratecontrol.set_minimum_delay((CRITIQUEBRAINZ_HOST, CRITIQUEBRAINZ_PORT), 50)


def result_review(album, metadata, data, reply, error):
    if error:
        album._requests -= 1
        album._finalize_loading(None)
        return

    try:
        reviews = load_json(data).get("reviews")
        if reviews:
            for review in reviews:
                if "last_revision" in review:
                    ident = review["entity_type"].replace("_", "-") + "_review_" + review["published_on"] + "_" + review["user"]["display_name"] + "_" + review["language"]
                    if review["last_revision"]["text"] is not None:
                        review_text = review["last_revision"]["text"] + "\n\nEine Bewertung mit " + str(review["last_revision"]["rating"]) + " von 5 Sternen veröffentlich durch " + review["user"]["display_name"] + " am " + review["published_on"] + " lizensiert unter " + review["full_name"] + ".";
                        if "comment:" + ident in metadata:
                            metadata["comment:" + ident] = review_text
                        else:
                            metadata.add("comment:" + ident, review_text)
                    review_url = "https://critiquebrainz.org/review/" + review["id"]
                    if "website:" + ident in metadata:
                        metadata["website:" + ident] = review_url
                    else:
                        metadata.add("website:" + ident, review_url)
            log.debug("%s: success parsing %s reviews (%s) from response", PLUGIN_NAME, reviews["count"], metadata["albumtitle"])
    except Exception as e:
        log.error("%s: error parsing review (%s) from response: %s", PLUGIN_NAME, metadata["albumtitle"], str(e))
    finally:
        album._requests -= 1
        album._finalize_loading(None)

def process_releasegroup(album, metadata, release):
    queryargs = {
        "entity_id": metadata["musicbrainz_releasegroupid"],
        "entity_type": "release_group"
    }
    album.tagger.webservice.get(
        CRITIQUEBRAINZ_HOST,
        CRITIQUEBRAINZ_PORT,
        "/ws/1/review",
        partial(result_review, album, metadata),
        priority=True,
        parse_response_type=None,
        queryargs=queryargs
    )
    album._requests += 1

def process_recording(album, metadata, track, release):
    queryargs = {
        "entity_id": metadata["musicbrainz_recordingid"],
        "entity_type": "recording"
    }
    album.tagger.webservice.get(
        CRITIQUEBRAINZ_HOST,
        CRITIQUEBRAINZ_PORT,
        "/ws/1/review",
        partial(result_review, album, metadata),
        priority=True,
        parse_response_type=None,
        queryargs=queryargs
    )
    album._requests += 1


register_album_metadata_processor(process_releasegroup)
register_track_metadata_processor(process_recording)




# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Release tag aggregation functions'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = ('Add functions to aggregate tags on a release:'
                      '<ul>'
                      '<li>$album_all(name)</li>'
                      '<li>‎$album_avg(name, precision=2)</li>'
                      '<li>‎$album_max(name, precision=2)</li>'
                      '<li>‎$album_min(name, precision=2)</li>'
                      '<li>‎$album_mode(name)</li>'
                      '<li>‎$album_distinct(name, separator=; )</li>'
                      '<li>‎$album_multi_avg(name, precision=2)</li>'
                      '<li>‎$album_multi_max(name, precision=2)</li>'
                      '<li>‎$album_multi_min(name, precision=2)</li>'
                      '<li>‎$album_multi_mode(name)</li>'
                      '<li>‎$album_multi_distinct(name, separator=; )</li>'
                      '</ul>'
                      '<b>The functions work only in file naming scripts and '
                      'the files should either be part of a release or cluster!</b>')
PLUGIN_VERSION = "0.4"
PLUGIN_API_VERSIONS = ["2.5", "2.6"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from functools import partial

from picard.cluster import Cluster
from picard.metadata import MULTI_VALUED_JOINER
from picard.script import script_function
from picard.script.parser import normalize_tagname


def get_parent_release(file):
    if file.parent:
        if isinstance(file.parent, Cluster):
            return file.parent
        elif hasattr(file.parent, 'album') and file.parent.album:
            return file.parent.album
    return None


def iter_release_values(name, file):
    parent = get_parent_release(file)
    if parent:
        for file in parent.iterfiles(save=True):
            yield file.metadata.get(name, '')
    else:
        yield file.metadata.get(name, '')


def iter_release_values_multi(name, file):
    parent = get_parent_release(file)
    if parent:
        for file in parent.iterfiles(save=True):
            yield from file.metadata.getall(name)
    else:
        yield from file.metadata.getall(name)


def try_iter_numeric(values, skip_non_numeric=False):
    for value in values:
        # First try treating the value as an integer, if this fails try float
        try:
            yield int(value)
        except (TypeError, ValueError):
            try:
                yield float(value)
            except (TypeError, ValueError):
                if not skip_non_numeric:
                    yield value


def aggregate_release_tags(parser, name, aggregate_func, multi=False):
    name = normalize_tagname(name)
    file = parser.file
    if file:
        iter_func = iter_release_values_multi if multi else iter_release_values
        return aggregate_func(iter_func(name, file))
    else:  # Nothing to aggregate, return base value
        return parser.context.get(name, '')


def format_number(value, precision=2):
    try:
        precision = int(precision)
    except (TypeError, ValueError):
        precision = 0
    if isinstance(value, int):
        return str(value)
    elif isinstance(value, float):
        fmt = '{:.' + str(precision) + 'f}'
        return fmt.format(value)
    elif not value:
        return ''
    else:
        return str(value)


def mode(values):
    """Returns the mode, the value with the most occurrences, from values."""
    l = list(values)
    if not l:
        return ''
    return max(set(l), key=l.count)


def average(values, precision=2):
    """Returns the arithmetic average of all numeric elements in values.

    Non-numeric elements are ignored.
    """
    numbers = list(try_iter_numeric(values, skip_non_numeric=True))
    if not numbers:
        return ''
    return format_number(sum(numbers) / len(numbers), precision=precision)


def natsort_min(values, precision=2):
    """Returns the smallest value from values, treats numeric strings as numbers."""
    return format_number(max(try_iter_numeric(values)), precision=precision)


def natsort_max(values, precision=2):
    """Returns the largest value from values, treats numeric strings as numbers."""
    return format_number(max(try_iter_numeric(values)), precision=precision)


def distinct(values, separator=MULTI_VALUED_JOINER):
    return separator.join(set(values))


@script_function(documentation="""`$album_all(name)`

Returns the value of the tag name if all files on the album have the same
value for this tag. Otherwise returns empty.
**Only works in File Naming scripts.**""")
def func_album_all(parser, name):
    def aggregate_func(values):
        common_value = ''
        for value in values:
            if not common_value:
                common_value = value
            if common_value != value:
                return ''
        return common_value

    return aggregate_release_tags(parser, name, aggregate_func)


@script_function(documentation="""`$album_mode(name)`

Returns the value with the most occurrences over all the files on the release for the given tag.
**Only works in File Naming scripts.**""")
def func_album_mode(parser, name):
    return aggregate_release_tags(parser, name, mode)


@script_function(documentation="""`$album_multi_mode(name)`

Returns the value with the most occurrences over all the files on the release for the given tag.
Similar to $album_mode(), but considers each value of multi-value tags separately.
**Only works in File Naming scripts.**""")
def func_album_multi_mode(parser, name):
    return aggregate_release_tags(parser, name, mode, multi=True)


@script_function(documentation="""`$album_min(name, precision=2)`

Returns the minimum value of all the files on the release for the given tag.
**Only works in File Naming scripts.**""")
def func_album_min(parser, name, precision="2"):
    aggregate_func = partial(natsort_min, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func)


@script_function(documentation="""`$album_multi_min(name, precision=2)`

Returns the minimum value of all the files on the release for the given tag.
Similar to $album_min(), but considers each value of multi-value tags separately.
**Only works in File Naming scripts.**""")
def releasetag_multi_min(parser, name, precision="2"):
    aggregate_func = partial(natsort_min, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func, multi=True)


@script_function(documentation="""`$album_max(name, precision=2)`

Returns the maximum value of all the files on the release for the given tag.
**Only works in File Naming scripts.**""")
def func_album_max(parser, name, precision="2"):
    aggregate_func = partial(natsort_max, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func)


@script_function(documentation="""`$album_multi_max(name, precision=2)`

Returns the maximum value of all the files on the release for the given tag.
Similar to $album_max(), but considers each value of multi-value tags separately.
**Only works in File Naming scripts.**""")
def releasetag_multi_max(parser, name, precision="2"):
    aggregate_func = partial(natsort_max, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func, multi=True)


@script_function(documentation="""`$album_avg(name, precision=2)`

Returns the arithmetical average value of all the files on the release for the given tag.
Non-numeric values are ignored. Returns empty if no numeric value is present.
**Only works in File Naming scripts.**""")
def func_album_avg(parser, name, precision="2"):
    aggregate_func = partial(average, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func)


@script_function(documentation="""`$album_multi_avg(name, precision=2)`

Returns the arithmetical average value of all the files on the release for the given tag.
Non-numeric values are ignored. Returns empty if no numeric value is present.
Similar to $album_avg(), but considers each value of multi-value tags separately.
**Only works in File Naming scripts.**""")
def func_album_multi_avg(parser, name, precision="2"):
    aggregate_func = partial(average, precision=precision)
    return aggregate_release_tags(parser, name, aggregate_func, multi=True)


@script_function(documentation="""`$album_distinct(name, separator=; )`

Returns a multi-value tag with all distinct values of tag across all the files on the release.
**Only works in File Naming scripts.**""")
def func_album_distinct(parser, name, separator=MULTI_VALUED_JOINER):
    aggregate_func = partial(distinct, separator=separator)
    return aggregate_release_tags(parser, name, aggregate_func)


@script_function(documentation="""`$album_multi_distinct(name, separator=; )`

Returns a multi-value tag with all distinct values of tag across all the files on the release.
Similar to $album_distinct(), but considers each value of multi-value tags separately.
**Only works in File Naming scripts.**""")
def func_album_multi_distinct(parser, name, separator=MULTI_VALUED_JOINER):
    aggregate_func = partial(distinct, separator=separator)
    return aggregate_release_tags(parser, name, aggregate_func, multi=True)



# -*- coding: utf-8 -*-
# Copyright © 2016 Daniel sobey <dns@dns.id.au >

# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See http://www.wtfpl.net/ for more details.

PLUGIN_NAME = 'Wikidata Genre'
PLUGIN_AUTHOR = 'Daniel Sobey, Sambhav Kothari'
PLUGIN_DESCRIPTION = 'Query wikidata to get genre tags'
PLUGIN_VERSION = '1.4.5'
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2"]
PLUGIN_LICENSE = 'WTFPL'
PLUGIN_LICENSE_URL = 'http://www.wtfpl.net/'

import re
from functools import partial
from picard import config, log
from picard.metadata import register_track_metadata_processor
from picard.plugins.wikidata.ui_options_wikidata import Ui_WikidataOptionsPage
from picard.ui.options import register_options_page, OptionsPage
from picard.webservice import ratecontrol


WIKIDATA_HOST = 'www.wikidata.org'
WIKIDATA_PORT = 443

ratecontrol.set_minimum_delay((WIKIDATA_HOST, WIKIDATA_PORT), 0)


def parse_ignored_tags(ignore_tags_setting):
    ignore_tags = []
    for tag in ignore_tags_setting.lower().split(','):
        if not tag:
            break
        tag = tag.strip()
        if tag.startswith('/') and tag.endswith('/'):
            try:
                tag = re.compile(tag[1:-1])
            except re.error:
                log.error(
                    'Error parsing ignored tag "%s"', tag, exc_info=True)
        ignore_tags.append(tag)
    return ignore_tags


def matches_ignored(ignore_tags, tag):
    if ignore_tags:
        tag = tag.lower().strip()
        for pattern in ignore_tags:
            if hasattr(pattern, 'match'):
                match = pattern.match(tag)
            else:
                match = pattern == tag
            if match:
                return True
    return False


class Wikidata:

    RELEASE_GROUP = 1
    ARTIST = 2
    WORK = 3

    def __init__(self):
        # Key: mbid, value: List of metadata entries to be updated when we have parsed everything
        self.requests = {}

        # Key: mbid, value: List of items to track the number of outstanding requests
        self.itemAlbums = {}

        # cache, items that have been found
        # key: mbid, value: list of strings containing the genre's
        self.cache = {}

        # metabrainz url
        self.mb_host = ''
        self.mb_port = ''

        # web service & logger
        self.ws = None
        self.log = None

        # settings from options, options
        self.use_release_group_genres = False
        self.use_artist_genres = False
        self.use_artist_only_if_no_release = False
        self.ignore_genres_from_these_artists = ''
        self.ignore_genres_from_these_artists_list = []
        self.use_work_genres = True
        self.ignore_these_genres = ''
        self.ignore_these_genres_list = []
        self.genre_delimiter = ''

    # not used
    def process_release(self, album, metadata, release):
        self.ws = album.tagger.webservice
        self.log = album.log
        item_id = metadata.getall('musicbrainz_releasegroupid')[0]

        log.info('WIKIDATA: Processing release group %s ' % item_id)
        self.process_request(metadata, album, item_id, item_type='release-group')
        for artist in metadata.getall('musicbrainz_albumartistid'):
            item_id = artist
            log.info('WIKIDATA: Processing release artist %s' % item_id)
            self.process_request(metadata, album, item_id, item_type='artist')

    # Main processing function
    # First see if we have already found what we need in the cache, finalize loading
    # Next see if we are already looking for the item
    #   If we are, add this item to the list of items to be updated once we find what we are looking for.
    #   Otherwise we are the first one to look up this item, start a new request
    # metadata, map containing the new metadata
    #
    def process_request(self, metadata, album, item_id, item_type):
        log.debug('WIKIDATA: Looking up cache for item: %s' % item_id)
        log.debug('WIKIDATA: Album request count: %s' % album._requests)
        log.debug('WIKIDATA: Item type %s' % item_type)
        if item_id in self.cache:
            log.debug('WIKIDATA: Found item in cache')
            genre_list = self.cache[item_id]
            new_genre = set(metadata.getall("genre"))
            new_genre.update(genre_list)
            #sort the new genre list so that they don't appear as new entries (not a change) next time
            metadata["genre"] = self.genre_delimiter.join(sorted(new_genre))
            return
        else:
            # pending requests are handled by adding the metadata object to a
            # list of things to be updated when the genre is found
            if item_id in self.itemAlbums:
                log.debug(
                    'WIKIDATA: Request already pending, add it to the list of items to update once this has been'
                    'found')
                self.requests[item_id].append(metadata)
            else:
                self.requests[item_id] = [metadata]
                self.itemAlbums[item_id] = album
                album._requests += 1

                log.debug('WIKIDATA: First request for this item')
                log.debug('WIKIDATA: About to call Musicbrainz to look up %s ' % item_id)

                path = '/ws/2/%s/%s' % (item_type, item_id)
                queryargs = {"inc": "url-rels"}

                self.ws.get(self.mb_host, self.mb_port, path, partial(self.musicbrainz_release_lookup, item_id,
                                                                      metadata),
                            parse_response_type="xml", priority=False, important=False, queryargs=queryargs)

    def musicbrainz_release_lookup(self, item_id, metadata, response, reply, error):
        found = False
        if error:
            log.error('WIKIDATA: Error retrieving release group info')
        else:
            if 'metadata' in response.children:
                if 'release_group' in response.metadata[0].children and self.use_release_group_genres:
                    if 'relation_list' in response.metadata[0].release_group[0].children:
                        for relation in response.metadata[0].release_group[0].relation_list[0].relation:
                            if relation.type == 'wikidata' and 'target' in relation.children:
                                found = True
                                wikidata_url = relation.target[0].text
                                log.debug('WIKIDATA: wikidata url found for RELEASE_GROUP: %s ', wikidata_url)
                                self.process_wikidata(Wikidata.RELEASE_GROUP, wikidata_url, item_id)
                if 'artist' in response.metadata[0].children and self.use_artist_genres:
                    if 'relation_list' in response.metadata[0].artist[0].children:
                        for relation in response.metadata[0].artist[0].relation_list[0].relation:
                            if relation.type == 'wikidata' and 'target' in relation.children:
                                found = True
                                wikidata_url = relation.target[0].text
                                self.process_wikidata(Wikidata.ARTIST, wikidata_url, item_id)
                                log.debug('WIKIDATA: wikidata url found for ARTIST: %s ', wikidata_url)
                if 'work' in response.metadata[0].children and self.use_work_genres:
                    if 'relation_list' in response.metadata[0].work[0].children:
                        for relation in response.metadata[0].work[0].relation_list[0].relation:
                            if relation.type == 'wikidata' and 'target' in relation.children:
                                found = True
                                wikidata_url = relation.target[0].text
                                log.debug('WIKIDATA: wikidata url found for WORK: %s ', wikidata_url)
                                self.process_wikidata(Wikidata.WORK, wikidata_url, item_id)
        if not found:
            log.debug('WIKIDATA: No wikidata url found for item_id: %s ', item_id)

        album = self.itemAlbums[item_id]
        album._requests -= 1
        if not album._requests:
            self.itemAlbums = {k: v for k, v in self.itemAlbums.items() if v != album}
            album._finalize_loading(None)
        log.info('WIKIDATA: Total remaining requests: %s' % album._requests)
        if not self.itemAlbums:
            self.requests.clear()
            log.info('WIKIDATA: Finished (A)')

    def process_wikidata(self, genre_source_type, wikidata_url, item_id):
        album = self.itemAlbums[item_id]
        album._requests += 1
        item = wikidata_url.split('/')[4]
        path = "/wiki/Special:EntityData/" + item + ".rdf"
        log.debug('WIKIDATA: Fetching from wikidata.org%s' % path)
        self.ws.get(WIKIDATA_HOST, WIKIDATA_PORT, path,
                    partial(self.parse_wikidata_response, item, item_id, genre_source_type),
                    parse_response_type="xml", priority=False, important=False)

    def parse_wikidata_response(self, item, item_id, genre_source_type, response, reply, error):
        genre_entries = []
        genre_list = []
        if error:
            log.error('WIKIDATA: error getting data from wikidata.org')
        else:
            if 'RDF' in response.children:
                node = response.RDF[0]
                for node1 in node.Description:
                    if 'about' in node1.attribs:
                        if node1.attribs.get('about') == 'http://www.wikidata.org/entity/%s' % item:
                            for key, val in list(node1.children.items()):
                                if key == 'P136':
                                    for i in val:
                                        if 'resource' in i.attribs:
                                            tmp = i.attribs.get('resource')
                                            if 'entity' == tmp.split('/')[3] and len(tmp.split('/')) == 5:
                                                genre_id = tmp.split('/')[4]
                                                log.debug(
                                                    'WIKIDATA: Found the wikidata id for the genre: %s' % genre_id)
                                                genre_entries.append(tmp)
                        else:
                            for tmp in genre_entries:
                                if tmp == node1.attribs.get('about'):
                                    list1 = node1.children.get('name')
                                    if not list1:
                                    	log.warning('WIKIDATA: Response does not contain a name field')
                                    else:
                                        for node2 in list1:
                                            if node2.attribs.get('lang') == 'en':
                                                genre = node2.text.title()
                                                if not matches_ignored(self.ignore_these_genres_list, genre):
                                                    genre_list.append(genre)
                                                    log.debug('New genre has been found and ALLOWED: %s' % genre)
                                                else:
                                                    log.debug('New genre has been found, but IGNORED: %s' % genre)

        if len(genre_list) > 0:
            log.debug('WIKIDATA: item_id: %s' % item_id)
            log.debug('WIKIDATA: Final list of wikidata id found: %s' % genre_entries)
            log.debug('WIKIDATA: Final list of genre: %s' % genre_list)
            log.info('WIKIDATA: Total items to update: %d ' % len(self.requests[item_id]))

            for metadata in self.requests[item_id]:
                if genre_source_type == Wikidata.RELEASE_GROUP:
                    metadata['~release_group_genre_sourced'] = True
                elif genre_source_type == Wikidata.ARTIST:
                    if self.use_artist_only_if_no_release and metadata['~release_group_genre_sourced'] or \
                            matches_ignored(self.ignore_genres_from_these_artists_list, metadata.get("artist")):
                        if item_id not in self.cache:
                            self.cache[item_id] = []
                        log.debug('WIKIDATA: NOT setting Artist-sourced genre: %s ' % genre_list)
                        continue
                    else:
                        log.debug('WIKIDATA: Setting Artist-sourced genre: %s ' % genre_list)

                # getall doesn't handle delimiters so we need to check-n-parse here
                old_genre_metadata = metadata.getall("genre")
                old_genre_list = []
                for genre in old_genre_metadata:
                    if self.genre_delimiter and self.genre_delimiter in genre:
                        old_genre_list.extend(genre.split(self.genre_delimiter))
                    else:
                        old_genre_list.append(genre)

                new_genre = set(old_genre_list)
                new_genre.update(genre_list)
                # Sort the new genre list so that they don't appear as new entries (not a change) next time
                log.debug('WIKIDATA: setting metadata genre to : %s ' % new_genre)
                if self.genre_delimiter:
                    metadata["genre"] = self.genre_delimiter.join(sorted(new_genre))
                else:
                    metadata["genre"] = sorted(new_genre)

                log.debug('WIKIDATA: setting cache genre to : %s ' % genre_list)
                self.cache[item_id] = genre_list
        else:
            log.debug('WIKIDATA: genre not found in wikidata')

        log.debug('WIKIDATA: seeing if we can finalize tags...')

        album = self.itemAlbums[item_id]
        album._requests -= 1
        if not album._requests:
            self.itemAlbums = {k: v for k, v in self.itemAlbums.items() if v != album}
            album._finalize_loading(None)
        log.info('WIKIDATA: total remaining requests: %s' % album._requests)
        if not self.itemAlbums:
            self.requests.clear()
            log.info('WIKIDATA: Finished (B)')

    def process_track(self, album, metadata, track, release):
        self.update_settings()
        self.ws = album.tagger.webservice
        self.log = album.log

        log.info('WIKIDATA: Processing Track...')
        if self.use_release_group_genres:
            for release_group in metadata.getall('musicbrainz_releasegroupid'):
                log.debug('WIKIDATA: Looking up release group metadata for %s ' % release_group)
                self.process_request(metadata, album, release_group, item_type='release-group')

        if self.use_artist_genres:
            for artist in metadata.getall('musicbrainz_albumartistid'):
                log.debug('WIKIDATA: Processing release artist %s' % artist)
                self.process_request(metadata, album, artist, item_type='artist')

        if self.use_artist_genres:
            for artist in metadata.getall('musicbrainz_artistid'):
                log.debug('WIKIDATA: Processing track artist %s' % artist)
                self.process_request(metadata, album, artist, item_type='artist')

        if self.use_work_genres:
            for workid in metadata.getall('musicbrainz_workid'):
                log.debug('WIKIDATA: Processing track artist %s' % workid)
                self.process_request(metadata, album, workid, item_type='work')

    def update_settings(self):
        self.mb_host = config.setting["server_host"]
        self.mb_port = config.setting["server_port"]
        self.use_release_group_genres = config.setting[""]
        self.use_work_genres = config.setting["wikidata_use_work_genres"]
        # Some changed settings could invalidate the cache, so clear it to be safe
        if self.use_release_group_genres != config.setting["wikidata_use_release_group_genres"]:
            self.use_release_group_genres = config.setting["wikidata_use_release_group_genres"]
            self.cache.clear()
        if self.use_artist_genres != config.setting["wikidata_use_artist_genres"]:
            self.use_artist_genres = config.setting["wikidata_use_artist_genres"]
            self.cache.clear()
        if self.use_artist_only_if_no_release != config.setting["wikidata_use_artist_only_if_no_release"]:
            self.use_artist_only_if_no_release = config.setting["wikidata_use_artist_only_if_no_release"]
            self.cache.clear()
        if self.ignore_genres_from_these_artists != parse_ignored_tags(
                config.setting["wikidata_ignore_genres_from_these_artists"]):
            self.ignore_genres_from_these_artists_list = parse_ignored_tags(
                config.setting["wikidata_ignore_genres_from_these_artists"])
            self.cache.clear()
        if self.use_work_genres != config.setting["wikidata_use_work_genres"]:
            self.use_work_genres = config.setting["wikidata_use_work_genres"]
            self.cache.clear()
        if self.ignore_these_genres != parse_ignored_tags(config.setting["wikidata_ignore_these_genres"]):
            self.ignore_these_genres = parse_ignored_tags(config.setting["wikidata_ignore_these_genres"])
            self.ignore_these_genres_list = parse_ignored_tags(
                config.setting["wikidata_ignore_these_genres"])
            self.cache.clear()
        if config.setting["write_id3v23"]:
            self.genre_delimiter = config.setting["wikidata_genre_delimiter"]


class WikidataOptionsPage(OptionsPage):
    NAME = "wikidata"
    TITLE = "Wikidata Genre"
    PARENT = "plugins"

    options = [
        config.BoolOption("setting", "wikidata_use_release_group_genres", True),
        config.BoolOption("setting", "wikidata_use_artist_genres", True),
        config.BoolOption("setting", "wikidata_use_artist_only_if_no_release", True),
        config.TextOption("setting", "wikidata_ignore_genres_from_these_artists", ""),
        config.BoolOption("setting", "wikidata_use_work_genres", True),
        config.TextOption("setting", "wikidata_ignore_these_genres", "seen live, favorites, /\\d+ of \\d+ stars/"),
        config.TextOption("setting", "wikidata_genre_delimiter", "; "),
    ]

    def __init__(self, parent=None):
        super(WikidataOptionsPage, self).__init__(parent)
        self.ui = Ui_WikidataOptionsPage()
        self.ui.setupUi(self)
        if not config.setting["write_id3v23"]:
            self.ui.genre_delimiter.setEnabled(False);
            self.ui.genre_delimiter_label.setEnabled(False);
        else:
            self.ui.genre_delimiter.setEnabled(True);
            self.ui.genre_delimiter_label.setEnabled(True);

    def load(self):
        setting = config.setting
        self.ui.use_release_group_genres.setChecked(setting["wikidata_use_release_group_genres"])
        self.ui.use_artist_genres.setChecked(setting["wikidata_use_artist_genres"])
        self.ui.use_artist_only_if_no_release.setChecked(setting["wikidata_use_artist_only_if_no_release"])
        self.ui.ignore_genres_from_these_artists.setText(setting["wikidata_ignore_genres_from_these_artists"])
        self.ui.use_work_genres.setChecked(setting["wikidata_use_work_genres"])
        self.ui.ignore_these_genres.setText(setting["wikidata_ignore_these_genres"])
        if config.setting["write_id3v23"]:
            self.ui.genre_delimiter.setEditText(setting["wikidata_genre_delimiter"])

    def save(self):
        setting = config.setting
        setting["wikidata_use_release_group_genres"] = self.ui.use_release_group_genres.isChecked()
        setting["wikidata_use_artist_genres"] = self.ui.use_artist_genres.isChecked()
        setting["wikidata_use_artist_only_if_no_release"] = self.ui.use_artist_only_if_no_release.isChecked()
        setting["wikidata_ignore_genres_from_these_artists"] = str(self.ui.ignore_genres_from_these_artists.text())
        setting["wikidata_use_work_genres"] = self.ui.use_work_genres.isChecked()
        setting["wikidata_ignore_these_genres"] = str(self.ui.ignore_these_genres.text())
        if config.setting["write_id3v23"]:
            setting["wikidata_genre_delimiter"] = str(self.ui.genre_delimiter.currentText())


wikidata = Wikidata()
register_track_metadata_processor(wikidata.process_track)
register_options_page(WikidataOptionsPage)



import doctest
import os
import glob
import json
import shutil
import tempfile
import unittest
from contextlib import redirect_stdout
from io import StringIO
from generate import build_json, zip_files


class GenerateTestCase(unittest.TestCase):

    """Run tests"""

    # The file that contains json data
    PLUGIN_FILE = "plugins.json"

    # The directory which contains plugin files
    PLUGIN_DIR = "plugins"

    def setUp(self):

        # Destination directory
        self.dest_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.dest_dir)
        self.plugin_file = os.path.join(self.dest_dir, self.PLUGIN_FILE)

    @staticmethod
    def with_suppressed_stdout(func, *args, **kwargs):
        out = StringIO()
        try:
            with redirect_stdout(out):
                func(*args, **kwargs)
        except Exception:
            print(out.getvalue())
            raise

    def test_generate_json(self):
        """
        Generates the json data from all the plugins
        and asserts that all plugins are accounted for.
        """

        self.with_suppressed_stdout(build_json, self.dest_dir)

        # Load the json file
        with open(self.plugin_file, "r") as in_file:
            plugin_json = json.load(in_file)["plugins"]

        # All top level directories in plugin_dir
        plugin_folders = next(os.walk(self.PLUGIN_DIR))[1]

        # Number of entries in the json should be equal to the
        # number of folders in plugin_dir
        self.assertEqual(len(plugin_json), len(plugin_folders))

    def test_generate_zip(self):
        """
        Generates zip files for all folders and asserts
        that all folders are accounted for.
        """

        self.with_suppressed_stdout(zip_files, self.dest_dir)

        # All zip files in plugin_dir
        plugin_zips = glob.glob(os.path.join(self.dest_dir, "*.zip"))

        # All top level directories in plugin_dir
        plugin_folders = next(os.walk(self.PLUGIN_DIR))[1]

        # Number of folders should be equal to number of zips
        self.assertEqual(len(plugin_zips), len(plugin_folders))

    def test_valid_json(self):
        """
        Asserts that the json data contains all the fields
        for all the plugins.
        """

        self.with_suppressed_stdout(build_json, self.dest_dir)

        # Load the json file
        with open(self.plugin_file, "r") as in_file:
            plugin_json = json.load(in_file)["plugins"]

        # All plugins should contain all required fields
        for module_name, data in plugin_json.items():
            self.assertIsInstance(data['name'], str)
            self.assertIsInstance(data['api_versions'], list)
            self.assertIsInstance(data['author'], str)
            self.assertIsInstance(data['description'], str)
            self.assertIsInstance(data['version'], str)



# -*- coding: utf-8

"""
Search longest common substrings using generalized suffix trees built with Ukkonen's algorithm

Author: Ilya Stepanov <code at ilyastepanov.com>

(c) 2013

Modified by <a href="https://github.com/MetaTunes">Mark Evens</a> as part of Picard Classical Extras project
Not for stand-alone use - use the original code
Accepts list or string inputs, but returns list outputs
Changed to allow a range of different special characters in case $ is in a string
(c) 2018
"""

import sys

END_OF_STRING = sys.maxsize

class SuffixTreeNode:
    """
    Suffix tree node class. Actually, it also respresents a tree edge that points to this node.
    """
    new_identifier = 0

    def __init__(self, start=0, end=END_OF_STRING):
        self.identifier = SuffixTreeNode.new_identifier
        SuffixTreeNode.new_identifier += 1

        # suffix link is required by Ukkonen's algorithm
        self.suffix_link = None

        # child edges/nodes, each dict key represents the first letter of an edge
        self.edges = {}

        # stores reference to parent
        self.parent = None

        # bit vector shows to which strings this node belongs
        self.bit_vector = 0

        # edge info: start index and end index
        self.start = start
        self.end = end

    def add_child(self, key, start, end):
        """
        Create a new child node

        Agrs:
            key: a char that will be used during active edge searching
            start, end: node's edge start and end indices

        Returns:
            created child node

        """
        child = SuffixTreeNode(start=start, end=end)
        child.parent = self
        self.edges[key] = child
        return child

    def add_exisiting_node_as_child(self, key, node):
        """
        Add an existing node as a child

        Args:
            key: a char that will be used during active edge searching
            node: a node that will be added as a child
        """
        node.parent = self
        self.edges[key] = node

    def get_edge_length(self, current_index):
        """
        Get length of an edge that points to this node

        Args:
            current_index: index of current processing symbol (usefull for leaf nodes that have "infinity" end index)
        """
        return min(self.end, current_index + 1) - self.start

    def __str__(self):
        return 'id=' + str(self.identifier)


class SuffixTree:
    """
    Generalized suffix tree
    """

    def __init__(self):
        # the root node
        self.root = SuffixTreeNode()

        # all strings are concatenaited together. Tree's nodes stores only indices
        self.input_string = []

        # number of strings stored by this tree
        self.strings_count = 0

        # list of tree leaves
        self.leaves = []

    def append_string(self, input_string, special_char):
        """
        Add new string to the suffix tree
        """
        start_index = len(self.input_string)
        current_string_index = self.strings_count

        # each sting should have a unique ending
        input_string += special_char + str(current_string_index)

        # gathering 'em all together
        self.input_string += input_string
        self.strings_count += 1

        # these 3 variables represents current "active point"
        active_node = self.root
        active_edge = 0
        active_length = 0

        # shows how many
        remainder = 0

        # new leaves appended to tree
        new_leaves = []

        # main circle
        for index in range(start_index, len(self.input_string)):
            previous_node = None
            remainder += 1
            while remainder > 0:
                if active_length == 0:
                    active_edge = index

                if self.input_string[active_edge] not in active_node.edges:
                    # no edge starting with current char, so creating a new leaf node
                    leaf_node = active_node.add_child(self.input_string[active_edge], index, END_OF_STRING)

                    # a leaf node will always be leaf node belonging to only one string
                    # (because each string has different termination)
                    leaf_node.bit_vector = 1 << current_string_index
                    new_leaves.append(leaf_node)

                    # doing suffix link magic
                    if previous_node is not None:
                        previous_node.suffix_link = active_node
                    previous_node = active_node
                else:
                    # ok, we've got an active edge
                    next_node = active_node.edges[self.input_string[active_edge]]

                    # walking down through edges (if active_length is bigger than edge length)
                    next_edge_length = next_node.get_edge_length(index)
                    if active_length >= next_node.get_edge_length(index):
                        active_edge += next_edge_length
                        active_length -= next_edge_length
                        active_node = next_node
                        continue

                    # current edge already contains the suffix we need to insert.
                    # Increase the active_length and go forward
                    if self.input_string[next_node.start + active_length] == self.input_string[index]:
                        active_length += 1
                        if previous_node is not None:
                            previous_node.suffix_link = active_node
                        previous_node = active_node
                        break

                    # splitting edge
                    split_node = active_node.add_child(
                        self.input_string[active_edge],
                        next_node.start,
                        next_node.start + active_length
                    )
                    next_node.start += active_length
                    split_node.add_exisiting_node_as_child(self.input_string[next_node.start], next_node)
                    leaf_node = split_node.add_child(self.input_string[index], index, END_OF_STRING)
                    leaf_node.bit_vector = 1 << current_string_index
                    new_leaves.append(leaf_node)

                    # suffix link magic again
                    if previous_node is not None:
                        previous_node.suffix_link = split_node
                    previous_node = split_node

                remainder -= 1

                # follow suffix link (if exists) or go to root
                if active_node == self.root and active_length > 0:
                    active_length -= 1
                    active_edge = index - remainder + 1
                else:
                    active_node = active_node.suffix_link if active_node.suffix_link is not None else self.root

        # update leaves ends from "infinity" to actual string end
        for leaf in new_leaves:
            leaf.end = len(self.input_string)
        self.leaves.extend(new_leaves)

    def find_longest_common_substrings(self, special_char):
        """
        Search longest common substrings in the tree by locating lowest common ancestors that belong to all strings
        """

        # all bits are set
        success_bit_vector = 2 ** self.strings_count - 1

        lowest_common_ancestors = []

        # going up to the root
        for leaf in self.leaves:
            node = leaf
            while node.parent is not None:
                if node.bit_vector != success_bit_vector:
                    # updating parent's bit vector
                    node.parent.bit_vector |= node.bit_vector
                    node = node.parent
                else:
                    # hey, we've found a lowest common ancestor!
                    lowest_common_ancestors.append(node)
                    break

        longest_common_substrings = []
        longest_length = 0

        # need to filter the result array and get the longest common strings
        for common_ancestor in lowest_common_ancestors:
            common_substring = []
            node = common_ancestor
            while node.parent is not None:
                label = self.input_string[node.start:node.end]
                common_substring = label + common_substring
                node = node.parent
            # remove unique endings (<special_char><number>), we don't need them anymore ...
            if special_char in common_substring:
                common_substring = common_substring[:common_substring.index(special_char)]
            # ... also in input strings to avoid mutation problems
            if len(common_substring) > longest_length:
                longest_length = len(common_substring)
                longest_common_substrings = [common_substring]
            elif len(common_substring) == longest_length and common_substring not in longest_common_substrings:
                longest_common_substrings.append(common_substring)

        return longest_common_substrings


def multi_lcs(strings_list):
    """
    Returns longest common string (or list) for a list of strings (or lists)
    :param strings_list: a list of lists or a list of strings
    :return: a list of longest common strings (or lists)
    (more than one is possible if they are distinct and of the same length)
    """

    if not isinstance(strings_list, list):
        return {'response': [], 'error': 'Argument is not a list'}
    arg_type = type(strings_list[0])
    for item in strings_list:
        if not isinstance(item, arg_type):
            return {'response': [], 'error': 'List members are not of the same type'}
    if arg_type is not list and arg_type is not str:
        return {'response': [], 'error': 'List members are not lists or strings'}

    suffix_tree = SuffixTree()
    special_char = None
    for char in ['|', '$', '#', '%', '@', '_']:
        test_strings = [x for x in strings_list if char in x]
        if not test_strings:
            special_char = char
            break
    if special_char:
        # print special_char
        for s in strings_list:
            if arg_type is list:
                s_copy = s[:]   # to avoid mutating input lists
            else:
                s_copy = s
            suffix_tree.append_string(s_copy, special_char)
        lcs = suffix_tree.find_longest_common_substrings(special_char)
    else:
        return {'response': [], 'error': 'Too many special characters'}
    if arg_type is list:
        if lcs:
            return {'response': lcs[0]}
        else:
            return {'response': [], 'error': 'Internal suffix tree problems'}
    else:
        if lcs:
            return {'response': [''.join(x) for x in lcs[0]]}
        else:
            return {'response': [], 'error': 'Internal suffix tree problems'}




# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2019, 2021-2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

PLUGIN_NAME = 'Work & Movement'
PLUGIN_AUTHOR = 'Philipp Wolfer'
PLUGIN_DESCRIPTION = '''Set work and movement based on work relationships.

This plugin attempts to set the `movement` and `movementnumber` tags, but only
if the work linked to the recording is part of a larger work. The `work` tag
then gets set to the work of which the movement is a part of.

If the recording is only linked to a simple work without separate parts then it
is not considered a proper work and the work and movement related tags will be
cleared.

The plugin will always set the original values, as loaded from MusicBrainz, of
the `work` and `musicbrainz_workid` tags into the variables `%%recording_work%`
and `%%recording_workid%` to be used in scripting.
'''
PLUGIN_VERSION = '1.1'
PLUGIN_API_VERSIONS = ['2.1', '2.2', '2.3', '2.4', '2.5', '2.6', '2.7', '2.8']
PLUGIN_LICENSE = 'GPL-2.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl-2.0.html'


import re

from .roman import (
    fromRoman,
    RomanError,
)

from picard import log
from picard.metadata import register_track_metadata_processor


_re_work_title = re.compile(r'(?P<work>.*):\s+(?P<movementnumber>[IVXLCDM]+)\.\s[\s,:;./]*(?P<movement>.*)')
_re_part_number = re.compile(r'(?P<number>[0-9IVXLCDM]+)\.?\s[\s,:;./]*')
_re_separators = re.compile(r'^[\s,:;./-]+')


class Work:
    def __init__(self, title, mbid=None):
        self.mbid = mbid
        self.title = title
        self.is_movement = False
        self.is_work = False
        self.part_number = 0
        self.parent = None

    def __str__(self):
        s = []
        if self.parent:
            s.append(str(self.parent))
        if self.is_movement:
            work_type = 'Movement'
        elif self.is_work:
            work_type = 'Work'
        else:
            work_type = 'Unknown'
        s.append('%s %i: %s' % (work_type, self.part_number, self.title))
        return '\n'.join(s)


def is_performance_work(rel):
    return (rel['target-type'] == 'work'
            and rel['direction'] == 'forward'
            and rel['type'] == 'performance')


def is_parent_work(rel):
    return (rel['target-type'] == 'work'
            and rel['direction'] == 'backward'
            and rel['type'] == 'parts')


def is_movement_like(rel):
    return ('movement' in rel['attributes']
            or 'act' in rel['attributes']
            or 'ordering-key' in rel)


def is_child_work(rel):
    return (rel['target-type'] == 'work'
            and rel['direction'] == 'forward'
            and rel['type'] == 'parts')


def number_to_int(s):
    """
    Converts a numeric string to int. `s` can also be a Roman numeral.
    """
    try:
        return int(s)
    except ValueError:
        try:
            return fromRoman(s)
        except RomanError as e:
            raise ValueError(e)


def parse_work_name(title):
    return _re_work_title.search(title)


def create_work_and_movement_from_title(work):
    """
    Attempts to parse work.title in the form "<Work>: <Number>. <Movement>",
    where <Number> is in Roman numerals.
    Sets the `is_movement` and `part_number` properties on `work` and creates
    a `parent` work if not already present.
    """
    title = work.title
    match = parse_work_name(title)
    if match:
        work.title = match.group('movement')
        work.is_movement = True
        try:
            number = number_to_int(match.group('movementnumber'))
        except ValueError as e:
            log.error(e)
            number = 0
        if not work.part_number:
            work.part_number = number
        elif work.part_number != number:
            log.warning('Movement number mismatch for "%s": %s != %i',
                        title, match.group('movementnumber'), work.part_number)
        if not work.parent:
            work.parent = Work(match.group('work'))
            work.parent.is_work = True
        elif work.parent.title != match.group('work'):
            log.warning('Movement work name mismatch for "%s": "%s" != "%s"',
                        title, match.group('work'), work.parent.title)
    return work


def normalize_movement_title(work):
    """
    Removes the parent work title and part number from the beginning of
    `work.title`. This ensures movement names don't contain duplicated
    information even if they do not follow the strict naming format used by
    `create_work_and_movement_from_title`.
    """
    movement_title = work.title
    if work.parent:
        work_title = work.parent.title
        if movement_title.startswith(work_title):
            movement_title = movement_title[len(work_title):]
            movement_title = _re_separators.sub('', movement_title)
    match = _re_part_number.match(movement_title)
    if match:
        # Only remove the number if it matches the part_number
        try:
            number = number_to_int(match.group('number'))
            if number == work.part_number:
                movement_title = _re_part_number.sub('', movement_title)
        except ValueError as e:
            log.warning(e)
    return movement_title


def parse_work(work_rel):
    work = Work(work_rel['title'], work_rel['id'])
    if 'relations' in work_rel:
        for rel in work_rel['relations']:
            # If this work has parents and is linked to those as 'movement' or
            # 'act' we consider it a part of a larger work and store it
            # in the movement tag. The parent will be set as the work.
            if is_parent_work(rel):
                if is_movement_like(rel):
                    work.is_movement = True
                    work.part_number = rel.get('ordering-key')
                    if 'work' in rel:
                        work.parent = parse_work(rel['work'])
                        work.parent.is_work = True
                    work.title = normalize_movement_title(work)
                else:
                    # Not a movement, but still part of a larger work.
                    # Mark it as a work.
                    work.is_work = True
            # If this work has any parts, we consider it a proper work.
            # This is a recording directly linked to a larger work.
            if is_child_work(rel):
                work.is_work = True
    return work


def unset_work(metadata):
    metadata.delete('work')
    metadata.delete('musicbrainz_workid')
    metadata.delete('movement')
    metadata.delete('movementnumber')
    metadata.delete('movementtotal')
    metadata.delete('showmovement')


def set_work(metadata, work):
    metadata['work'] = work.title
    metadata['musicbrainz_workid'] = work.mbid
    metadata['showmovement'] = 1


def process_track(album, metadata, track, release):
    if 'recording' in track:
        recording = track['recording']
    else:
        recording = track

    if 'relations' not in recording:
        return

    # Remember original work and work ID
    metadata['~recording_work'] = metadata.getall('work')
    metadata['~recording_workid'] = metadata.getall('musicbrainz_workid')

    # Try to find a suitable work linked to the recording.
    # As a fallback try to get work + movement information by parsing
    # the recording title.
    work = Work(recording['title'])
    for rel in recording['relations']:
        if is_performance_work(rel):
            work = parse_work(rel['work'])
            # Only use the first work that qualifies as a work or movement
            log.debug('Found work:\n%s', work)
            if work.is_movement or work.is_work:
                break

    unset_work(metadata)
    if not work.is_movement:
        work = create_work_and_movement_from_title(work)

    if work.is_movement and work.parent and work.parent.is_work:
        metadata['movement'] = work.title
        if work.part_number:
            metadata['movementnumber'] = work.part_number
        set_work(metadata, work.parent)
    elif work.is_work:
        set_work(metadata, work)


register_track_metadata_processor(process_track)



# -*- coding: utf-8 -*-

PLUGIN_NAME = u"Compatible TXXX frames"
PLUGIN_AUTHOR = u'Tungol'
PLUGIN_DESCRIPTION = """This plugin improves the compatibility of ID3 tags \
by using only a single value for TXXX frames. Multiple value TXXX frames \
technically don't comply with the ID3 specification."""
PLUGIN_VERSION = "0.1"
PLUGIN_API_VERSIONS = ["2.0"]
PLUGIN_LICENSE = "GPL-2.0-or-later"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.html"

from picard import config
from picard.formats import register_format
from picard.formats.id3 import MP3File, TrueAudioFile, DSFFile, AiffFile
from mutagen import id3


id3v24_join_with = '; '


def build_compliant_TXXX(self, encoding, desc, values):
    """Return a TXXX frame with only a single value.

    Use id3v23_join_with as the sperator if using id3v2.3, otherwise the value
    set in this plugin (default "; ").
    """
    if config.setting['write_id3v23']:
        sep = config.setting['id3v23_join_with']
    else:
        sep = id3v24_join_with
    joined_values = [sep.join(values)]
    return id3.TXXX(encoding=encoding, desc=desc, text=joined_values)


# I can't actually remove the original MP3File et al formats once they're
# registered. This depends on the name of the replacements sorting after the
# name of the originals, because picard.formats.guess_format picks the last
# item from a sorted list.


class MP3FileCompliant(MP3File):
    """Alternate MP3 format class which uses single-value TXXX frames."""

    build_TXXX = build_compliant_TXXX


class TrueAudioFileCompliant(TrueAudioFile):
    """Alternate TTA format class which uses single-value TXXX frames."""

    build_TXXX = build_compliant_TXXX


class DSFFileCompliant(DSFFile):
    """Alternate DSF format class which uses single-value TXXX frames."""

    build_TXXX = build_compliant_TXXX


class AiffFileCompliant(AiffFile):
    """Alternate AIFF format class which uses single-value TXXX frames."""

    build_TXXX = build_compliant_TXXX


register_format(MP3FileCompliant)
register_format(TrueAudioFileCompliant)
register_format(DSFFileCompliant)
register_format(AiffFileCompliant)



import json
from functools import partial
from typing import Callable, List, NamedTuple, Optional, TypeVar
from urllib.parse import urlsplit, urlunsplit

from picard.webservice import WebService
from PyQt5.QtCore import QByteArray
from PyQt5.QtNetwork import QNetworkReply

from . import obj

DEEZER_HOST = 'api.deezer.com'
DEEZER_PORT = 443


T = TypeVar('T', bound=obj.APIObject)
SearchCallback = Callable[[List[T], Optional[QNetworkReply.NetworkError]], None]
APIURLCallback = Callable[[Optional[T], Optional[QNetworkReply.NetworkError]], None]


class SearchOptions(NamedTuple('SearchOptions', [('artist', str), ('album', str), ('track', str), ('label', str)])):
    """
    Options for the advanced search.
    """

    def __str__(self):
        options = ['{}:"{}"'.format(k, v) for k, v in self._asdict().items() if v]
        return ' '.join(options)


# Python 3.5 cannot set defaults values in an othodox way.
SearchOptions.__new__.__defaults__ = ('',) * len(SearchOptions._fields)


class Client:
    def __init__(self, webservice: WebService):
        self.webservice = webservice
        self._get = partial(self.webservice.get, DEEZER_HOST, DEEZER_PORT)

    def advanced_search(self, options: SearchOptions, callback: SearchCallback[obj.APIObject]):
        path = '/search'

        def handler(document: QByteArray, _: QNetworkReply, error: Optional[QNetworkReply.NetworkError]):
            try:
                parsed_doc = json.loads(str(document, 'utf-8'))
            except json.JSONDecodeError:
                callback([], error)
            else:
                result = []
                error = None
                if 'data' in parsed_doc:
                    result = [obj.parse_json(dct) for dct in parsed_doc['data']]
                elif 'error' in parsed_doc:
                    error = parsed_doc['error'].get('message', 'Deezer responded with an unknown error')
                else:
                    error = 'Deezer returned an unexpected response'
                callback(result, error)

        self._get(path,
                  queryargs={'q': str(options)},
                  parse_response_type=None,
                  handler=handler)

    def obj_from_url(self, url: str, callback: APIURLCallback[obj.APIObject]):
        def handler(document: QByteArray, _: QNetworkReply, error: Optional[QNetworkReply.NetworkError]):
            try:
                deezer_obj = obj.parse_json(str(document, 'utf-8'))
            except json.JSONDecodeError:
                deezer_obj = None
            finally:
                callback(deezer_obj, error)

        self._get(self._remove_language_path(urlsplit(url).path),
                  parse_response_type=None,
                  handler=handler)

    @staticmethod
    def api_url(url: str) -> str:
        parsed = urlsplit(url)
        path = Client._remove_language_path(parsed.path)
        return urlunsplit((parsed.scheme, DEEZER_HOST, path, parsed.query, parsed.fragment))

    @staticmethod
    def _remove_language_path(path: str) -> str:
        # Deezer has a 2-letter language path, e.g. /us/track/123.
        lang_len = 2
        paths = path[1:].split('/', maxsplit=1)
        return path[lang_len + 1:] if len(paths[0]) == lang_len else path



from picard.config import TextOption, get_config
from typing import Optional

COLLECTION_ID = "add_to_collection_id"


def collection_id_option() -> TextOption:
    return TextOption(section="setting", name=COLLECTION_ID, default=None)


def collection_id() -> Optional[str]:
    config = get_config()
    if COLLECTION_ID in config.setting:
        return config.setting[COLLECTION_ID]
    return None


def set_collection_id(value: str) -> None:
    config = get_config()
    config.setting[COLLECTION_ID] = value



# -*- coding: utf-8 -*-
# AcousticBrainz plugin for Picard
#
# Copyright (C) 2021 Wargreen <wargreen@lebib.org>
# Copyright (C) 2021 Philipp Wolfer <ph.wolfer@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Plugin metadata
# =============================================================================

PLUGIN_NAME = 'AcousticBrainz Tags'
PLUGIN_AUTHOR = ('Wargreen <wargreen@lebib.org>, '
                 'Hugo Geoffroy "pistache" <pistache@lebib.org>, '
                 'Philipp Wolfer <ph.wolfer@gmail.com>, '
                 'Regorxxx <regorxxx@protonmail.com>')
PLUGIN_DESCRIPTION = '''
Tag files with tags from the AcousticBrainz database, all highlevel classifiers
and tonal/rhythm data.
<br/><br/>
By default, only simple mood and genre information is saved, but the plugin can
be configured to include all highlevel data.
<br/><br/>
Based on code from Andrew Cook, Sambhav Kothari
<br/><br/>
<b>WARNING:</b> Experimental plugin. All guarantees voided by use.'''

PLUGIN_LICENSE = "GPL-2.0"
PLUGIN_LICENSE_URL = "https://www.gnu.org/licenses/gpl-2.0.txt"
PLUGIN_VERSION = "2.2.3"
PLUGIN_API_VERSIONS = ["2.0", "2.1", "2.2", "2.3", "2.4", "2.5", "2.6", "2.7"]

# Plugin configuration
# =============================================================================

ACOUSTICBRAINZ_HOST = "acousticbrainz.org"
ACOUSTICBRAINZ_PORT = 80


# Subset of the low level data to add as tags.
# Represent the data as nested dicts.

SUBLOWLEVEL_SUBSET = {"rhythm": {"bpm": None},
                      "tonal": {"chords_changes_rate": None,
                                "chords_key": None,
                                "chords_scale": None,
                                "key_key": None,
                                "key_scale": None}}

# Imports
# =============================================================================

from functools import partial
from json import dumps as dump_json

from picard import (
    config,
    log,
)
from picard.metadata import (
    register_album_metadata_processor,
    register_track_metadata_processor,
)
from picard.ui.options import (
    OptionsPage,
    register_options_page,
)
from picard.webservice import ratecontrol
from picard.plugins.acousticbrainz.ui_options_acousticbrainz_tags import Ui_AcousticBrainzOptionsPage

ratecontrol.set_minimum_delay((ACOUSTICBRAINZ_HOST, ACOUSTICBRAINZ_PORT), 1000)

# Constants
# =============================================================================

LOWLEVEL = "lowlevel"
HIGHLEVEL = "highlevel"


# Logging utilities
# -------------------------------------------------------------------------

def log_msg(logger, text, *args):
    logger("%s: " + text, PLUGIN_NAME, *args)

def debug(*args):
    log_msg(log.debug, *args)

def warning(*args):
    log_msg(log.warning, *args)

def error(*args):
    log_msg(log.error, *args)


# TrackDataProcessor class
# =============================================================================
# (used to apply AcousticBrainz data to Track metadata)

class TrackDataProcessor:
    def __init__(self, recording_id, metadata, level, data, files=None):
        self.recording_id = recording_id
        self.metadata = metadata
        self.level = level
        self.data = self._extract_data(data)
        self.files = files
        self.do_simplemood = config.setting["acousticbrainz_add_simplemood"]
        self.simplemood_tagname = config.setting["acousticbrainz_simplemood_tagname"]
        self.do_simplegenre = config.setting["acousticbrainz_add_simplegenre"]
        self.simplegenre_tagname = config.setting["acousticbrainz_simplegenre_tagname"]
        self.do_keybpm = config.setting["acousticbrainz_add_keybpm"]
        self.do_fullhighlevel = config.setting["acousticbrainz_add_fullhighlevel"]
        self.do_sublowlevel = config.setting["acousticbrainz_add_sublowlevel"]

    # Logging utilities
    # -------------------------------------------------------------------------

    def log(self, logger, text, *args):
        log_msg(logger, "[%s: %s] " + text, self.shortid, self.title, *args)

    def debug(self, *args):
        self.log(log.debug, *args)

    def warning(self, *args):
        self.log(log.warning, *args)

    def error(self, *args):
        self.log(log.error, *args)

    # Read-only properties
    # -------------------------------------------------------------------------

    @property
    def shortid(self):
        return self.recording_id[:8]

    @property
    def title(self):
        return self.metadata["title"]

    # Callback
    # -------------------------------------------------------------------------

    def process(self):
        if not self.data:
            self.warning('No %s data for track %s', self.level, self.recording_id)

        if self.level == HIGHLEVEL:
            if self.do_simplemood:
                self.process_simplemood()
            if self.do_simplegenre:
                self.process_simplegenre()
            if self.do_fullhighlevel:
                self.process_fullhighlevel()

        if self.level == LOWLEVEL:
            if self.do_keybpm:
                self.process_keybpm()
            if self.do_sublowlevel:
                self.process_sublowlevel()

    # Processing helper methods
    # -------------------------------------------------------------------------

    def _extract_data(self, data):
        if not self.recording_id in data:
            return {}
        if self.level == LOWLEVEL:
            return data[self.recording_id]["0"]
        elif self.level == HIGHLEVEL:
            return data[self.recording_id]["0"]["highlevel"]

    def filter_data(self, data, subset):
        result = {}
        for key, value in subset.items():
            if key in data:
                if isinstance(value, dict):
                    self.debug("filter_data : traversing %s", key)
                    result[key] = self.filter_data(data[key], value)
                else:
                    self.debug("filter_data : adding result %s", key)
                    result[key] = data[key]
            else:
                self.debug("filter_data : Subset key %s not found in data", key)
                continue
        self.debug("filter_data : result : %s", result)
        return result

    def update_metadata(self, name, values):
        self.metadata[name] = values
        if self.files:
            for file in self.files:
                file.metadata[name] = values

    # Processing methods
    # -------------------------------------------------------------------------
    # (fill metadata with fetched data)

    def process_simplemood(self):
        self.debug("processing simplemood data")

        mood_tagname = self.simplemood_tagname;
        moods = []

        for classifier, data in self.data.items():
            if "value" in data:
                value = data["value"]
                inverted = value.startswith("not_")
                if classifier.startswith("mood_") and not inverted:
                    moods.append(value)

        self.update_metadata(mood_tagname, moods)

    def process_simplegenre(self):
        self.debug("processing simplegenre data")

        genre_tagname = self.simplegenre_tagname;
        genres = []

        for classifier, data in self.data.items():
            if "value" in data:
                value = data["value"]
                inverted = value.startswith("not_")
                if classifier.startswith("genre_") and not inverted:
                    genres.append(value)

        self.update_metadata(genre_tagname, genres)

    def process_fullhighlevel(self):
        self.debug("processing fullhighlevel data")

        f_count, c_count = 0, 0
        for classifier, data in self.data.items():
            classifier = classifier.lower()
            if "all" in data:
                c_count += 1
                for feature, proba in data["all"].items():
                    feature = feature.lower()
                    f_count += 1
                    self.update_metadata("ab:hi:{}:{}".format(classifier, feature), dump_json(proba))
            else:
                self.warning("fullhighlevel : ignored invalid classifier data (%s)", classifier)

        self.debug("fullhighlevel : parsed %d features from %d classifiers", f_count, c_count)

    def process_keybpm(self):
        self.debug("processing keybpm data")
        if "tonal" in self.data:
            tonal_data = self.data["tonal"]
            if "key_key" in tonal_data:
                key = tonal_data["key_key"]
                if "key_scale" in tonal_data:
                    if tonal_data["key_scale"] == "minor":
                        key += "m"
                self.update_metadata('key', key)
                self.debug("track '%s' is in key %s", self.title, key)

        if "rhythm" in self.data:
            rhythm_data = self.data["rhythm"]
            if "bpm" in rhythm_data:
                bpm = int(rhythm_data["bpm"] + 0.5)
                self.update_metadata('bpm', bpm)
                self.debug("keybpm : Track '%s' has %s bpm", self.title, bpm)

    def process_sublowlevel(self):
        self.debug("processing sublowlevel data")
        subset = SUBLOWLEVEL_SUBSET

        filtered_data = self.filter_data(self.data, subset)

        f_count, c_count = 0, 0
        for classifier, data in filtered_data.items():
            classifier = classifier.lower()
            c_count += 1
            for feature, proba in data.items():
                feature = feature.lower()
                f_count += 1
                self.update_metadata("ab:lo:{}:{}".format(classifier, feature), proba)

        self.debug("sublowlevel : parsed %d features from %d classifiers", f_count, c_count)


class AcousticBrainzRequest:

    MAX_BATCH_SIZE = 25

    def __init__(self, webservice, recording_ids):
        self.webservice = webservice
        self.recording_ids = recording_ids

    def request_highlevel(self, callback):
        self._batch('high-level', self.recording_ids, callback, {})

    def request_lowlevel(self, callback):
        self._batch('low-level', self.recording_ids, callback, {})

    def _batch(self, action, recording_ids, callback, result, response=None, reply=None, error=None):
        if response and not error:
            self._merge_results(result, response)

        if not recording_ids or error:
            callback(result, error)
            return

        batch = recording_ids[:self.MAX_BATCH_SIZE]
        recording_ids = recording_ids[self.MAX_BATCH_SIZE:]
        self._do_request(action, batch,
            callback=partial(self._batch, action, recording_ids, callback, result))

    def _do_request(self, action, recording_ids, callback):
        self.webservice.get(
            ACOUSTICBRAINZ_HOST,
            ACOUSTICBRAINZ_PORT,
            '/api/v1/%s' % action,
            callback,
            priority=True,
            parse_response_type='json',
            queryargs=self._get_query_args(action, recording_ids)
        )

    def _get_query_args(self, action, recording_ids):
        queryargs = {
            'recording_ids': ';'.join(recording_ids),
        }
        if action == 'high-level':
            queryargs['map_classes'] = 'true'
        return queryargs

    def _merge_results(self, full, new):
        mapping = new.get('mbid_mapping', {})
        new = {mapping.get(k, k): v for (k, v) in new.items() if k != 'mbid_mapping'}
        full.update(new)


# Plugin class
# =============================================================================
# (provides track processing callback)

class AcousticBrainzPlugin:

    result_cache = {
        LOWLEVEL: {},
        HIGHLEVEL: {},
    }

    def process_album(self, album, metadata, release):
        debug('Processing album %s', album.id)
        recording_ids = self.get_recording_ids(release)
        self.run_requests(album, recording_ids, self.album_callback)

    def process_track(self, album, metadata, track_node, release_node):
        # Run requests for standalone recordings
        if not release_node and 'id' in track_node:
            recording_id = track_node['id']
            debug('Processing recording %s', recording_id)
            self.run_requests(album, [recording_id], partial(self.nat_callback, recording_id))
        # Apply metadata changes for already loaded results for album tracks
        elif 'recording' in track_node:
            recording_id = track_node['recording']['id']
            for level in (LOWLEVEL, HIGHLEVEL):
                result = self.result_cache[level].get(album.id)
                if result:
                    self.apply_result(recording_id, metadata, level, result)

    def run_requests(self, album, recording_ids, callback):
        request = AcousticBrainzRequest(album.tagger.webservice, recording_ids)
        if self.do_highlevel:
            album._requests += 1
            request.request_highlevel(partial(callback, HIGHLEVEL, album))
        if self.do_lowlevel:
            album._requests += 1
            request.request_lowlevel(partial(callback, LOWLEVEL, album))

    @property
    def do_highlevel(self):
        return (config.setting["acousticbrainz_add_simplemood"]
            or config.setting["acousticbrainz_add_simplegenre"]
            or config.setting["acousticbrainz_add_fullhighlevel"])

    @property
    def do_lowlevel(self):
        return (config.setting["acousticbrainz_add_keybpm"]
            or config.setting["acousticbrainz_add_sublowlevel"])

    def get_recording_ids(self, release):
        return [track['recording']['id'] for track in self.iter_tracks(release)]

    @staticmethod
    def iter_tracks(release):
        for media in release['media']:
            if 'pregap' in media:
                yield media['pregap']

            if 'tracks' in media:
                yield from media['tracks']

            if 'data-tracks' in media:
                yield from media['data-tracks']

    def album_callback(self, level, album, result=None, error=None):
        if not error:
            # Store the result, the actual processing will be done by the
            # track metadata processor.
            self.result_cache[level][album.id] = result
            album.run_when_loaded(partial(self.clear_cache, level, album))
        album._requests -= 1
        album._finalize_loading(error)

    def nat_callback(self, recording_id, level, album, result=None, error=None):
        for track in album.tracks:
            if track.id == recording_id:
                self.apply_result(track.id, track.metadata, level, result, files=track.files)

    def apply_result(self, recording_id, metadata, level, result, files=None):
        debug('Updating recording %s with %s results', recording_id, level)
        processor = TrackDataProcessor(recording_id, metadata, level, result, files=files)
        processor.process()

    def clear_cache(self, level, album):
        try:
            del self.result_cache[level][album.id]
        except KeyError:
            pass


# Plugin options page
# =============================================================================
# (define plugin options and link with user interface)

class AcousticBrainzOptionsPage(OptionsPage):
    NAME = "acousticbrainz_tags"
    TITLE = "AcousticBrainz tags"
    PARENT = "plugins"

    options = [
        config.BoolOption("setting", "acousticbrainz_add_simplemood", True),
        config.TextOption("setting", "acousticbrainz_simplemood_tagname", "ab:mood"),
        config.BoolOption("setting", "acousticbrainz_add_simplegenre", True),
        config.TextOption("setting", "acousticbrainz_simplegenre_tagname", "ab:genre"),
        config.BoolOption("setting", "acousticbrainz_add_keybpm", False),
        config.BoolOption("setting", "acousticbrainz_add_fullhighlevel", False),
        config.BoolOption("setting", "acousticbrainz_add_sublowlevel", False)
    ]

    def __init__(self, parent=None):
        super(AcousticBrainzOptionsPage, self).__init__(parent)
        self.ui = Ui_AcousticBrainzOptionsPage()
        self.ui.setupUi(self)

    def load(self):
        setting = config.setting
        self.ui.add_simplemood.setChecked(setting["acousticbrainz_add_simplemood"])
        self.ui.simplemood_tagname.setText(setting["acousticbrainz_simplemood_tagname"])
        self.ui.add_simplegenre.setChecked(setting["acousticbrainz_add_simplegenre"])
        self.ui.simplegenre_tagname.setText(setting["acousticbrainz_simplegenre_tagname"])
        self.ui.add_fullhighlevel.setChecked(setting["acousticbrainz_add_fullhighlevel"])
        self.ui.add_keybpm.setChecked(setting["acousticbrainz_add_keybpm"])
        self.ui.add_sublowlevel.setChecked(setting["acousticbrainz_add_sublowlevel"])

    def save(self):
        setting = config.setting
        setting["acousticbrainz_add_simplemood"] = self.ui.add_simplemood.isChecked()
        setting["acousticbrainz_simplemood_tagname"] = str(self.ui.simplemood_tagname.text())
        setting["acousticbrainz_add_simplegenre"] = self.ui.add_simplegenre.isChecked()
        setting["acousticbrainz_simplegenre_tagname"] = str(self.ui.simplegenre_tagname.text())
        setting["acousticbrainz_add_keybpm"] = self.ui.add_keybpm.isChecked()
        setting["acousticbrainz_add_fullhighlevel"] = self.ui.add_fullhighlevel.isChecked()
        setting["acousticbrainz_add_sublowlevel"] = self.ui.add_sublowlevel.isChecked()


plugin = AcousticBrainzPlugin()
register_album_metadata_processor(plugin.process_album)
register_track_metadata_processor(plugin.process_track)
register_options_page(AcousticBrainzOptionsPage)



PLUGIN_NAME = 'Feat. Artists in Titles'
PLUGIN_AUTHOR = 'Lukas Lalinsky, Michael Wiencek, Bryan Toth, JeromyNix (NobahdiAtoll)'
PLUGIN_DESCRIPTION = 'Move "feat." from artist names to album and track titles. Match is case insensitive.'
PLUGIN_VERSION = "0.5"
PLUGIN_API_VERSIONS = ["0.9.0", "0.10", "0.15", "0.16", "2.0"]

from picard.metadata import register_album_metadata_processor, register_track_metadata_processor
import re

_feat_re = re.compile(r"([\s\S]+) feat\.([\s\S]+)", re.IGNORECASE)


def move_album_featartists(tagger, metadata, release):
    match = _feat_re.match(metadata["albumartist"])
    if match:
        metadata["albumartist"] = match.group(1)
        metadata["album"] += " (feat.%s)" % match.group(2)
    match = _feat_re.match(metadata["albumartistsort"])
    if match:
        metadata["albumartistsort"] = match.group(1)


def move_track_featartists(tagger, metadata, track, release):
    match = _feat_re.match(metadata["artist"])
    if match:
        metadata["artist"] = match.group(1)
        metadata["title"] += " (feat.%s)" % match.group(2)
    match = _feat_re.match(metadata["artistsort"])
    if match:
        metadata["artistsort"] = match.group(1)

register_album_metadata_processor(move_album_featartists)
register_track_metadata_processor(move_track_featartists)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'plugins/search_engine_lookup/ui_options_search_engine_editor.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SearchEngineEditorDialog(object):
    def setupUi(self, SearchEngineEditorDialog):
        SearchEngineEditorDialog.setObjectName("SearchEngineEditorDialog")
        SearchEngineEditorDialog.setWindowModality(QtCore.Qt.ApplicationModal)
        SearchEngineEditorDialog.resize(560, 205)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SearchEngineEditorDialog.sizePolicy().hasHeightForWidth())
        SearchEngineEditorDialog.setSizePolicy(sizePolicy)
        SearchEngineEditorDialog.setMinimumSize(QtCore.QSize(560, 205))
        SearchEngineEditorDialog.setMaximumSize(QtCore.QSize(560, 300))
        self.verticalLayout = QtWidgets.QVBoxLayout(SearchEngineEditorDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea = QtWidgets.QScrollArea(SearchEngineEditorDialog)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 542, 187))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_4 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_4.setWordWrap(True)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_2.addWidget(self.label_4)
        spacerItem = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(-1, -1, -1, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.le_title = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        self.le_title.setFont(font)
        self.le_title.setToolTipDuration(-1)
        self.le_title.setObjectName("le_title")
        self.gridLayout.addWidget(self.le_title, 0, 1, 1, 1)
        self.le_url = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(9)
        self.le_url.setFont(font)
        self.le_url.setObjectName("le_url")
        self.gridLayout.addWidget(self.le_url, 1, 1, 1, 1)
        self.img_valid_title = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.img_valid_title.setMinimumSize(QtCore.QSize(16, 16))
        self.img_valid_title.setText("")
        self.img_valid_title.setObjectName("img_valid_title")
        self.gridLayout.addWidget(self.img_valid_title, 0, 2, 1, 1)
        self.img_valid_url = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.img_valid_url.setMinimumSize(QtCore.QSize(16, 16))
        self.img_valid_url.setText("")
        self.img_valid_url.setObjectName("img_valid_url")
        self.gridLayout.addWidget(self.img_valid_url, 1, 2, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 5, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.pb_save = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.pb_save.setObjectName("pb_save")
        self.horizontalLayout.addWidget(self.pb_save)
        self.pb_cancel = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.pb_cancel.setObjectName("pb_cancel")
        self.horizontalLayout.addWidget(self.pb_cancel)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.retranslateUi(SearchEngineEditorDialog)
        QtCore.QMetaObject.connectSlotsByName(SearchEngineEditorDialog)

    def retranslateUi(self, SearchEngineEditorDialog):
        _translate = QtCore.QCoreApplication.translate
        SearchEngineEditorDialog.setWindowTitle(_translate("SearchEngineEditorDialog", "Edit Search Engine Provider"))
        self.label_4.setText(_translate("SearchEngineEditorDialog", "<html><head/><body><p>Enter the title and URL for the search engine provider. Titles must be at least two non-space characters long, and must not be the same as the title of an existing provider.</p><p>When entering the URL the macro <span style=\"font-weight:600;\">%search%</span> must be included. This will be replaced by the list of search words separated by plus signs when the url is sent to your browser for display.</p></body></html>"))
        self.label_2.setText(_translate("SearchEngineEditorDialog", "Title:"))
        self.label_3.setText(_translate("SearchEngineEditorDialog", "URL:"))
        self.le_title.setToolTip(_translate("SearchEngineEditorDialog", "The title to show in the list for the search engine provider"))
        self.le_url.setToolTip(_translate("SearchEngineEditorDialog", "The URL to use for the search engine provider"))
        self.pb_save.setText(_translate("SearchEngineEditorDialog", "Save"))
        self.pb_cancel.setText(_translate("SearchEngineEditorDialog", "Cancel"))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Bob Swift (rdswift)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# "Persistent variables display" context is based on the code from
# the "View Script Variables" plugin.

PLUGIN_NAME = 'Persistent Variables'
PLUGIN_AUTHOR = 'Bob Swift (rdswift)'
PLUGIN_DESCRIPTION = '''
<p>
This plugin provides the ability to store and retrieve script variables that persist across tracks and albums.
This allows things like finding and storing the earliest recording date of all of the tracks on an album.
</p><p>
There are two types of persistent variables maintained - album variables and session variables. Album variables
persist across all tracks on an album.  Each album's information is stored separately, and is reset when the
album is refreshed. The information is cleared when an album is removed.  Session variables persist across all
albums and tracks, and are cleared when Picard is shut down or restarted.
</p><p>
This plugin adds eight new scripting functions to allow management of persistent script variables:
<ul>
<li>$set_a(name,value) : Sets the album persistent variable name to value.</li>
<li>$unset_a(name) : Unsets the album persistent variable name.</li>
<li>$get_a(name) : Gets the album persistent variable name.</li>
<li>$clear_a() : Clears all album persistent variables.</li>
<li>$set_s(name,value) : Sets the session persistent variable name to value.</li>
<li>$unset_s(name) : Unsets the session persistent variable name.</li>
<li>$get_s(name) : Gets the session persistent variable name.</li>
<li>$clear_s() : Clears all session persistent variables.</li>
</ul>
</p><p>
Please see the <a href="https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/persistent_variables/docs/README.md">user guide</a> on GitHub for more information.
</p>
'''
PLUGIN_VERSION = '1.1'
PLUGIN_API_VERSIONS = ['2.0', '2.1', '2.2', '2.3', '2.4', '2.6', '2.7']
PLUGIN_LICENSE = 'GPL-2.0-or-later'
PLUGIN_LICENSE_URL = 'https://www.gnu.org/licenses/gpl-2.0.html'

PLUGIN_USER_GUIDE_URL = 'https://github.com/rdswift/picard-plugins/blob/2.0_RDS_Plugins/plugins/persistent_variables/docs/README.md'

from PyQt5 import QtWidgets

from picard import log
from picard.album import (
    Album,
    register_album_post_removal_processor,
)
from picard.file import File
from picard.metadata import register_album_metadata_processor
from picard.plugin import PluginPriority
from picard.plugins.persistent_variables.ui_variables_dialog import (
    Ui_VariablesDialog,
)
from picard.script import register_script_function
from picard.script.parser import normalize_tagname
from picard.track import Track

from picard.ui.itemviews import (
    BaseAction,
    register_album_action,
    register_file_action,
    register_track_action,
)


class PersistentVariables:
    album_variables = {}
    session_variables = {}

    @classmethod
    def clear_album_vars(cls, album):
        if album:
            cls.album_variables[album] = {}

    @classmethod
    def set_album_var(cls, album, key, value):
        if album:
            if album not in cls.album_variables:
                cls.clear_album_vars(album)
            if key:
                cls.album_variables[album][key] = value

    @classmethod
    def unset_album_var(cls, album, key):
        if album and album in cls.album_variables:
            cls.album_variables[album].pop(key, None)

    @classmethod
    def unset_album_dict(cls, album):
        if album:
            cls.album_variables.pop(album, None)

    @classmethod
    def get_album_var(cls, album, key):
        if album in cls.album_variables:
            return cls.album_variables[album][key] if key in cls.album_variables[album] else ""
        return ""

    @classmethod
    def clear_session_vars(cls):
        cls.session_variables = {}

    @classmethod
    def set_session_var(cls, key, value):
        if key:
            cls.session_variables[key] = value

    @classmethod
    def unset_session_var(cls, key):
        cls.session_variables.pop(key, None)

    @classmethod
    def get_session_var(cls, key):
        return cls.session_variables[key] if key in cls.session_variables else ""

    @classmethod
    def get_album_dict(cls, album):
        if album and album in cls.album_variables:
            return cls.album_variables[album]
        return {}

    @classmethod
    def get_session_dict(cls):
        return cls.session_variables


def _get_album_id(parser):
    file = parser.file
    if file:
        if file.parent and hasattr(file.parent, 'album') and file.parent.album:
            return str(file.parent.album.id)
        else:
            return ""
    # Fall back to parser context to allow processing on albums newly retrieved from MusicBrainz
    return parser.context['musicbrainz_albumid']


def func_set_s(parser, name, value):
    if value:
        PersistentVariables.set_session_var(normalize_tagname(name), value)
    else:
        func_unset_s(parser, name)
    return ""


def func_unset_s(parser, name):
    PersistentVariables.unset_session_var(normalize_tagname(name))
    return ""


def func_get_s(parser, name):
    return PersistentVariables.get_session_var(normalize_tagname(name))


def func_clear_s(parser):
    PersistentVariables.clear_session_vars()
    return ""


def func_unset_a(parser, name):
    album_id = _get_album_id(parser)
    log.debug("{0}: Unsetting album '{1}' variable '{2}'".format(PLUGIN_NAME, album_id, normalize_tagname(name),))
    if album_id:
        PersistentVariables.unset_album_var(album_id, normalize_tagname(name))
    return ""


def func_set_a(parser, name, value):
    album_id = _get_album_id(parser)
    log.debug("{0}: Setting album '{1}' persistent variable '{2}' to '{3}'".format(PLUGIN_NAME, album_id, normalize_tagname(name), value,))
    if album_id:
        PersistentVariables.set_album_var(album_id, normalize_tagname(name), value)
    return ""


def func_get_a(parser, name):
    album_id = _get_album_id(parser)
    log.debug("{0}: Getting album '{1}' persistent variable '{2}'".format(PLUGIN_NAME, album_id, normalize_tagname(name),))
    if album_id:
        return PersistentVariables.get_album_var(album_id, normalize_tagname(name))
    return ""


def func_clear_a(parser):
    album_id = _get_album_id(parser)
    log.debug("{0}: Clearing album '{1}' persistent variables dictionary".format(PLUGIN_NAME, album_id,))
    if album_id:
        PersistentVariables.clear_album_vars(album_id)
    return ""


def initialize_album_dict(album, album_metadata, release_metadata):
    album_id = str(album.id)
    log.debug("{0}: Initializing album '{1}' persistent variables dictionary".format(PLUGIN_NAME, album_id,))
    PersistentVariables.clear_album_vars(album_id)


def destroy_album_dict(album):
    album_id = str(album.id)
    log.debug("{0}: Destroying album '{1}' persistent variables dictionary".format(PLUGIN_NAME, album_id,))
    PersistentVariables.unset_album_dict(album_id)


class ViewVariables(BaseAction):
    NAME = 'View persistent variables'

    def callback(self, objs):
        obj = objs[0]
        files = self.tagger.get_files_from_objects(objs)
        if files:
            obj = files[0]
        dialog = ViewVariablesDialog(obj)
        dialog.exec_()


class ViewVariablesDialog(QtWidgets.QDialog):

    def __init__(self, obj, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_VariablesDialog()
        self.ui.setupUi(self)
        self.ui.buttonBox.accepted.connect(self.accept)
        self.album_id = ""
        self.setWindowTitle("Persistent Variables")
        if isinstance(obj, Album):
            self.album_id = str(obj.id)
        if isinstance(obj, File):
            if obj.parent and hasattr(obj.parent, 'album') and obj.parent.album:
                self.album_id = str(obj.parent.album.id)
        elif isinstance(obj, Track):
            if obj.album:
                self.album_id = str(obj.album.id)
        album_dict = PersistentVariables.get_album_dict(self.album_id)
        album_count = len(album_dict)
        session_dict = PersistentVariables.get_session_dict()
        session_count = len(session_dict)

        table = self.ui.metadata_table
        key_example, value_example = self.get_table_items(table, 0)
        self.key_flags = key_example.flags()
        self.value_flags = value_example.flags()
        table.setRowCount(album_count + session_count + 2)
        i = 0
        self.add_separator_row(table, i, "Album Variables", album_count)
        i += 1
        for key in sorted(album_dict.keys()):
            key_item, value_item = self.get_table_items(table, i)
            key_item.setText(key)
            value_item.setText(album_dict[key])
            i += 1
        self.add_separator_row(table, i, "Session Variables", session_count)
        i += 1
        for key in sorted(session_dict.keys()):
            key_item, value_item = self.get_table_items(table, i)
            key_item.setText(key)
            value_item.setText(session_dict[key])
            i += 1

    def add_separator_row(self, table, i, title, count):
        key_item, value_item = self.get_table_items(table, i)
        font = key_item.font()
        font.setBold(True)
        key_item.setFont(font)
        key_item.setText(title)
        value_item.setText("{0} item{1}".format(count, "" if count == 1 else "s",))

    def get_table_items(self, table, i):
        key_item = table.item(i, 0)
        value_item = table.item(i, 1)
        if not key_item:
            key_item = QtWidgets.QTableWidgetItem()
            key_item.setFlags(self.key_flags)
            table.setItem(i, 0, key_item)
        if not value_item:
            value_item = QtWidgets.QTableWidgetItem()
            value_item.setFlags(self.value_flags)
            table.setItem(i, 1, value_item)
        return key_item, value_item

viewer = ViewVariables()


# Register the new functions
register_script_function(func_set_a, name='set_a',
    documentation="""`$set_a(name,value)`

Sets the album persistent variable `name` to `value`.""")

register_script_function(func_unset_a, name='unset_a',
    documentation="""`$unset_a(name)`

Unsets the album persistent variable `name`.""")

register_script_function(func_get_a, name='get_a',
    documentation="""`$get_a(name)`

Gets the album persistent variable `name`.""")

register_script_function(func_clear_a, name='clear_a',
    documentation="""`$clear_a()`

Clears all album persistent variables.""")

register_script_function(func_set_s, name='set_s',
    documentation="""`$set_s(name,value)`

Sets the session persistent variable `name` to `value`.""")

register_script_function(func_unset_s, name='unset_s',
    documentation="""`$unset_s(name)`

Unsets the session persistent variable `name`.""")

register_script_function(func_get_s, name='get_s',
    documentation="""`$get_s(name)`

Gets the session persistent variable `name`.""")

register_script_function(func_clear_s, name='clear_s',
    documentation="""`$clear_s()`

Clears all session persistent variables.""")


# Register the processers
register_album_metadata_processor(initialize_album_dict, priority=PluginPriority.HIGH)
register_album_post_removal_processor(destroy_album_dict)


# Register context actions
register_file_action(viewer)
register_track_action(viewer)
register_album_action(viewer)



# -*- coding: utf-8 -*-

PLUGIN_NAME = "Copy Cluster to Clipboard"
PLUGIN_AUTHOR = "Michael Elsdörfer, Sambhav Kothari"
PLUGIN_DESCRIPTION = "Exports a cluster's tracks to the clipboard, so it can be copied into the tracklist field on MusicBrainz"
PLUGIN_VERSION = "1.0"
PLUGIN_API_VERSIONS = ["2.0"]


from PyQt5 import QtWidgets
from picard.cluster import Cluster
from picard.util import format_time
from picard.ui.itemviews import BaseAction, register_cluster_action


class CopyClusterToClipboard(BaseAction):
    NAME = "Copy Cluster to Clipboard..."

    def callback(self, objs):
        if len(objs) != 1 or not isinstance(objs[0], Cluster):
            return
        cluster = objs[0]

        artists = set()
        for i, file in enumerate(cluster.files):
            artists.add(file.metadata["artist"])

        tracks = []
        for i, file in enumerate(cluster.files):
            try:
                i = int(file.metadata["tracknumber"])
            except:
                i += 1

            if len(artists) > 1:
                tracks.append((i, "%s. %s - %s (%s)" % (
                    i,
                    file.metadata["title"],
                    file.metadata["artist"],
                    format_time(file.metadata.length))))
            else:
                tracks.append((i, "%s. %s (%s)" % (
                    i,
                    file.metadata["title"],
                    format_time(file.metadata.length))))

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText("\n".join([x[1] for x in sorted(tracks)]))


register_cluster_action(CopyClusterToClipboard())
